class PagedQueryResult:
    """
    defines results of a paged query.
    :ivar str direction: 'asc' or 'desc' direction.
    :ivar int page: page number return on current result.
    :ivar list results: list of results records return by the query.
    :ivar int size: number of records per page.
    :ivar str sort: column name used in sorting.
    :ivar int total: total number of results on all pages.
    """

    def __init__(self,
                 total: int = None,
                 page: int = None,
                 size: int = None,
                 sort: str = None,
                 direction: str = None):
        self.total = total
        self.page = page
        if page < 0:
            raise ValueError("page must be a positive number")
        self.size = size
        if size < 0 or size > 200:
            raise ValueError("size is limited to 200")
        self.sort = sort
        if sort is None:
            raise ValueError("sort is mandatory")
        self.direction = direction
        if direction != 'asc' and direction != 'desc':
            raise ValueError("direction must be asc or desc")
        self.results = None

    @classmethod
    def from_dict(cls, data: dict, dto_class=None) -> "PagedQueryResult":
        """
        build a PagedQueryResult from a JSON response dictionary.
        :param data: dictionary from the API response.
        :param dto_class: class with a from_dict classmethod used to deserialize each result item.
        :return: PagedQueryResult with deserialized results.
        """
        result = cls(
            total=int(data.get("total", 0)),
            page=data.get("page", 1),
            size=data.get("size", 50),
            sort=data.get("sort", "id"),
            direction=data.get("direction", "desc"),
        )
        result.results = []
        for obj in data.get("results", []):
            if dto_class is not None and hasattr(dto_class, 'from_dict'):
                result.results.append(dto_class.from_dict(obj))
            else:
                result.results.append(obj)
        return result

    def to_json(self):
        """
        format the results
        """

        if self.direction == "asc":
            sort_dir = "asc"
        else:
            sort_dir = "desc"

        formatted_results = []
        if self.results:
            for obj in self.results:
                if hasattr(obj, 'to_json'):
                    formatted_results.append(obj.to_json())
                elif hasattr(obj, 'to_dict'):
                    formatted_results.append(obj.to_dict())
                else:
                    formatted_results.append(obj)
        return {
            "page": self.page,
            "size": self.size,
            "sort": self.sort,
            "direction": sort_dir,
            "total": self.total,
            "results": formatted_results
        }


from .api_dto import ApiDto


class DataStore(ApiDto):

    @classmethod
    def route(cls):
        return "datastores"

    @classmethod
    def from_dict(cls, data):
        obj = DataStore()
        obj.from_json(data)
        return obj

    @classmethod
    def get_id_type(cls) -> type:
        return int

    def __init__(self,
                 data_store_id: int = None,
                 name: str = None,
                 bucket: str = None,
                 measurement: str = None,
                 retention_seconds: int = None,
                 do_not_use_sensor_id: bool = False):
        self.data_store_id = data_store_id
        self.name = name
        self.bucket = bucket
        self.measurement = measurement
        self.retention_seconds = retention_seconds
        self.do_not_use_sensor_id = do_not_use_sensor_id

    def api_id(self) -> str:
        return str(self.data_store_id).upper()

    def endpoint(self) -> str:
        return "DataStores"

    def set_id(self, id_value):
        if not isinstance(id_value, int):
            raise TypeError(f"data store id must be an integer")
        self.data_store_id = id_value

    def to_json(self, target: str = None):
        obj = {}
        if self.data_store_id is not None:
            obj["id"] = self.data_store_id
        if self.name is not None:
            obj["name"] = self.name
        if self.bucket is not None:
            obj["bucket"] = self.bucket
        if self.measurement is not None:
            obj["measurement"] = self.measurement
        if self.retention_seconds is not None:
            obj["retentionSeconds"] = int(self.retention_seconds)
        if self.do_not_use_sensor_id is not None:
            obj["doNotUseSensorId"] = self.do_not_use_sensor_id
        return obj

    def from_json(self, obj):
        if "id" in obj.keys():
            self.data_store_id = int(obj["id"])
        if "name" in obj.keys():
            self.name = str(obj['name'])
        if "bucket" in obj.keys():
            self.bucket = str(obj['bucket'])
        if "measurement" in obj.keys():
            self.measurement = str(obj['measurement'])
        if "retentionSeconds" in obj.keys():
            self.retention_seconds = int(obj['retentionSeconds'])
        if "doNotUseSensorId" in obj.keys():
            if isinstance(obj["doNotUseSensorId"], str):
                if obj["doNotUseSensorId"] == "False" or obj["doNotUseSensorId"] is None:
                    self.do_not_use_sensor_id = False
                else:
                    self.do_not_use_sensor_id = True
            else:
                self.do_not_use_sensor_id = obj['doNotUseSensorId']



import json
import uuid

import dill
import os
import types
from .api_dto import ApiDto
import inspect
from collections import OrderedDict


def get_imports(line) -> dict:
    keywords = []
    for item in line.split(' '):
        keywords.extend(item.split(','))
    keywords[:] = [x for x in keywords if x]

    if len(keywords) <= 1:
        return {}

    i = 0
    current_key = None
    definition = {}
    while i < len(keywords):
        word = keywords[i]
        if word in ['from', 'as', 'import']:
            current_key = word
        elif current_key is not None:
            if current_key not in definition or definition[current_key] is None:
                definition[current_key] = [word]
            else:
                definition[current_key].append(word)
        else:
            raise ValueError(f'{line} cannot be parsed for possible import.')
        i = i+1

    return definition


class ScriptConfig(ApiDto):
    """
    a script config defines execution properties for a specific script.
    usually to define how a pipeline should call your script.

    :ivar str function: name of function referencing the script.
    :ivar dict properties: specific properties to configure the script.
    :ivar dict properties_mapping: mapping used to rename properties from context to script expected name.
    """

    def __init__(self,
                 function=None,
                 library=None,
                 properties_mapping=None,
                 properties: dict = None):
        self.function = function
        self.library = library
        self.properties = properties
        self.properties_mapping = properties_mapping

    def from_json(self, obj):
        if isinstance(obj, str):
            self.function = obj
            return
        else:
            if "function" in obj:
                self.function = obj["function"]
            if "library" in obj:
                self.library = obj["library"]
            elif "script" in obj:
                self.function = obj["script"]
            if "properties" in obj:
                self.properties = obj["properties"]
            if "properties_mapping" in obj:
                self.properties_mapping = obj["properties_mapping"]

    def to_json(self, target: str = None):
        obj = {
            "function": self.function
        }
        if self.library is not None:
            obj["library"] = self.library
        if self.properties is not None and isinstance(self.properties, dict):
            obj["properties"] = self.properties
        if self.properties_mapping is not None and isinstance(self.properties_mapping, dict):
            obj["properties_mapping"] = self.properties_mapping
        return obj


class Script(ApiDto):
    """
    a piece of python code that can either transform data, generate plot, train Machine Learning models.

    :ivar uuid.UUID script_id: technical id of the script
    :ivar str name: name of your script function, same as the code object.
    :ivar str description: provide an insight-full description of what does your script.
    :ivar callable() function: your function as a callable function - 'your function without ()'.
    :ivar str source: source code extracted from your source file.
    """

    @classmethod
    def route(cls):
        return "scripts"

    @classmethod
    def from_dict(cls, data):
        obj = Script()
        obj.from_json(data)
        return obj

    @classmethod
    def get_type(cls):
        return "dill"

    def __init__(self, script_id=None, description=None, function=None,
                 exact_names=False, exact_numbers=False):

        # Id
        if script_id is None:
            script_id = uuid.uuid4()
        self.script_id = script_id
        self._name = None

        # Properties
        self.description = description
        self.needExactColumnNumbers = exact_numbers
        self.needExactColumnNames = exact_names

        # Validation Properties
        self.canGeneratePlot = False
        self.canGenerateModel = False
        self.canGenerateData = False
        self.status = "draft"
        self.inputColumns = []
        self.outputColumns = []

        # Source code property
        self.source = None

        # Function properties (code)
        self.function = None
        if function is not None:
            self.copy(function)

    @property
    def name(self):
        """
        Name of the function
        :return: name of the function
        """
        return self._name

    def api_id(self) -> str:
        """
        Id of the script (script_id)

        :return: string formatted UUID of the script.
        """
        return str(self.script_id).upper()

    def endpoint(self) -> str:
        """
        Name of the endpoints used to manipulate scripts.
        :return: Endpoint name.
        """
        return "Scripts"

    def to_json(self, target: str = None):
        """
        Convert the script to a dictionary compatible to JSON format.

        :return: dictionary representation of the Script object.
        """
        obj = {
            "id": str(self.script_id),
            "canGeneratePlot": str(self.canGeneratePlot),
            "canGenerateModel": str(self.canGenerateModel),
            "canGenerateData": str(self.canGenerateData),
            "status": str(self.status),
            "needExactColumnNumbers": str(self.needExactColumnNumbers),
            "needExactColumnNames": str(self.needExactColumnNames),
            "inputColumns": json.dumps(list(self.inputColumns)),
            "outputColumns": json.dumps(list(self.outputColumns))
        }
        if self.name is not None:
            obj["name"] = str(self.name)
        if self.name is not None:
            obj["description"] = str(self.description)
        return obj

    def from_json(self, obj):
        """
        Load the Script entity from a dictionary representation of the Script.

        :param obj: Dict version of the Script.
        """
        if "id" in obj.keys():
            self.script_id = uuid.UUID(obj["id"])
        if "name" in obj.keys():
            self._name = obj["name"]
        if "description" in obj.keys():
            if obj["description"] != "None":
                self.description = obj["description"]
        if "canGeneratePlot" in obj.keys():
            if isinstance(obj["canGeneratePlot"], str) and obj["canGeneratePlot"].lower() == "false":
                self.canGeneratePlot = False
            else:
                self.canGeneratePlot = bool(obj["canGeneratePlot"])
        if "canGenerateModel" in obj.keys():
            if isinstance(obj["canGenerateModel"], str) and obj["canGenerateModel"].lower() == "false":
                self.canGenerateModel = False
            else:
                self.canGenerateModel = bool(obj["canGenerateModel"])
        if "canGenerateData" in obj.keys():
            if isinstance(obj["canGenerateData"], str) and obj["canGenerateData"].lower() == "false":
                self.canGenerateData = False
            else:
                self.canGenerateData = bool(obj["canGenerateData"])
        if "status" in obj.keys():
            self.status = str(obj["status"]).lower()
        if "needExactColumnNumbers" in obj.keys():
            if isinstance(obj["needExactColumnNumbers"], str) and obj["needExactColumnNumbers"].lower() == "false":
                self.needExactColumnNumbers = False
            else:
                self.needExactColumnNumbers = bool(obj["needExactColumnNumbers"])
        if "needExactColumnNames" in obj.keys():
            if isinstance(obj["needExactColumnNames"], str) and obj["needExactColumnNames"].lower() == "false":
                self.needExactColumnNames = False
            else:
                self.needExactColumnNames = bool(obj["needExactColumnNames"])
        if "inputColumns" in obj.keys():
            if obj["inputColumns"].lower() == "false":
                self.inputColumns = False
            else:
                self.inputColumns = json.loads(obj["inputColumns"])
        if "outputColumns" in obj.keys():
            if obj["outputColumns"].lower() == "false":
                self.outputColumns = False
            else:
                self.outputColumns = json.loads(obj["outputColumns"])

    def copy(self, myfunction):
        """
        Copy your function code and decorators to a format executable by the Wizata App.
        :param myfunction: your function - pass the function itself as parameter
        """

        if myfunction.__code__.co_argcount < 1:
            raise ValueError('your function must contains at least one parameter')

        self.function = Function()

        self.function.id = myfunction.__name__
        self.function.params = inspect.signature(myfunction).parameters

        self._name = myfunction.__name__

        self.function.code = myfunction.__code__

        f_globals = myfunction.__globals__
        self.function.globals = []

        # ADD GLOBAL IMPORTS
        k_global: str
        for k_global in f_globals:

            if self._name == k_global:
                pass
            elif isinstance(myfunction.__globals__[k_global], types.ModuleType):
                module = f_globals[k_global]
                self.function.append_global(definition={
                    "var": k_global,
                    "module": str(module.__name__)
                })

            elif not k_global.startswith('__') and not k_global.startswith('@') and not k_global.startswith('_'):
                obj = f_globals[k_global]
                if callable(obj) and hasattr(obj, '__module__'):
                    definition = {
                        "var": k_global,
                        'module': obj.__module__
                    }
                    if hasattr(obj, '__name__'):
                        definition['class'] = obj.__name__
                    self.function.append_global(definition)

        # ADD LOCAL IMPORTS
        for line in inspect.getsourcelines(myfunction)[0]:
            line = line.strip()
            if line.startswith("from ") or line.startswith("import "):
                imports = get_imports(line)
                if 'import' not in imports:
                    raise ValueError(f'from/import statement missing import {line}')

                if 'from' in imports:
                    if len(imports['from']) != 1:
                        raise ValueError(f'unsupported multiple from statement {line}')
                    if 'as' in imports:
                        if len(imports["import"]) != 1:
                            raise ValueError(f'unsupported multiple import statement with as {line}')
                        self.function.append_global({
                            'module': imports['from'][0],
                            'class': imports['import'][0],
                            'var': imports['as'][0]
                        })
                    else:
                        for import_class in imports['import']:
                            self.function.append_global({
                                'module': imports['from'][0],
                                'class': import_class,
                                'var': import_class
                            })
                else:
                    if 'as' in imports:
                        if len(imports["import"]) != 1:
                            raise ValueError(f'unsupported multiple import statement with as {line}')
                        self.function.append_global({
                            'module': imports['import'][0],
                            'var': imports['as'][0]
                        })
                    else:
                        for import_class in imports['import']:
                            self.function.append_global({
                                'module': import_class,
                                'var': import_class
                            })

        if myfunction.__code__.co_filename is not None:
            if os.path.exists(myfunction.__code__.co_filename):
                with open(myfunction.__code__.co_filename, "r") as f:
                    source_code = f.read()
                self.source = source_code


class Function:
    """
    Python Code Function

    :ivar id: technical name and then id of the function
    :ivar code: code of the function __code__
    :ivar globals: modules references __globals__
    """

    def __init__(self):
        self.id = None
        self.code = None
        self.globals = None
        self.params = OrderedDict()

    def append_global(self, definition: dict):
        exclusion_list = [
            '__main__',
            'IPython.core.interactiveshell',
            'IPython.core.autocall'
        ]
        if 'module' in definition and definition['module'] not in exclusion_list:
            self.globals.append(definition)





import uuid
from .solution_component import SolutionComponent
from .paged_query_result import PagedQueryResult
from .pipeline import AlertType
from .mlmodel import ModelInfo, ModelFile
from datetime import datetime
import pandas


class ApiInterface:
    """
    Interface with all methods available both to the pipeline runners and the Data Science API.
    """

    def __init__(self):
        self._dto_registry = {}

    def _set_registry(self, name: str, registry: dict):
        """
        internal call to set API registry definition.
        :param name: cloud_dsapi, context_dsapi, ...
        """
        if name not in ['cloud_dsapi', 'cloud_context']:
            raise ValueError(f'unsupported api type {name}')
        for entity_name in registry:
            if name in registry[entity_name] and len(registry[entity_name][name]) > 0:
                self._dto_registry[entity_name] = {
                    "class": registry[entity_name]["class"],
                    "operations": registry[entity_name][name]
                }

    def lists(self, entity):
        """
        lists all elements of a specific entity.
        :param entity: plural name of the entity or class (e.g. scripts, plots, mlmodels, dataframes...)
        :return: list of all elements with at least the id property.
        """
        pass

    def get(self,
            obj=None,
            id: uuid.UUID = None,
            key: str = None,
            entity=None):
        """
        get record content from DS API.

        - get will look first for obj then for id then for key (e.g. if id and key specified, key is ignored )

        :param obj: a wizata_dsapi.<Entity>() with UUID set - fetch using technical UUID

        :param id: a UUID of a specific object (use in combination of entity)
        :param key: a logical key of a specific object (e.g. hardware id for DataPoint and Twin (use in combination of entity)
        :param entity: plural name of the entity or class (use in combination of entity)(e.g. scripts, plots, mlmodels, dataframes...)
        :return: object with all properties or None if not found.
        """
        pass

    def create(self, obj):
        """
        create and save an object on the server
        :param obj: object from a supported entity
        :return: object created
        """
        pass

    def update(self, obj):
        """
        update and save an object on the server
        :param obj: object from a supported entity
        """
        pass

    def delete(self, obj):
        """
        delete an object on the server
        :param obj: object from a supported entity
        """
        pass

    def query(self,
              datapoints: list[str] = None,
              start: datetime = None,
              end: datetime = None,
              interval: int = None,
              agg_method: str = "mean",
              template: str = None,
              twin: str = None,
              null: str = None,
              filters: dict = None,
              options: dict = None,
              group: dict = None,
              field: str = None) -> pandas.DataFrame:
        """
        Query a dataframe from timeseries DB.
        :param agg_method:
        :param datapoints: list of datapoints to fetch.
        :param start: start datetime of range to fetch
        :param end: end datetime of range to fetch
        :param interval: interval in milliseconds.
        :param template: template to fetch.
        :param twin: hardware ID of twin to fetch based on template.
        :param null: default 'drop' will drop NaN values. If not intended behavior please set it to 'ignore' or 'all'.
        :param filters: dict of filters.
        :param options: dict of options.
        :param field: by default 'value' if none, can be used to retrieve 'eventId'
        :param group: can be used to set group system and event retrieving instructions.
        :return: dataframe
        """
        pass

    def _get_class(self,
                   entity,
                   operation: str = None):
        """
        get ApiDto class definition for a specific entity and valid that desired operation is supported.
        :param entity: str plural name of the entity or direct class definition.
        :param operation: optional - set operation name to verify it is valid on the class.
        """
        if isinstance(entity, str):
            if entity in self._dto_registry.keys():
                self._verify_operation(entity, operation)
                return self._dto_registry[entity]["class"]
            else:
                raise TypeError(f"Api entity {entity} is not supported.")
        else:
            for entity_name in self._dto_registry:
                if self._dto_registry[entity_name]["class"] == entity:
                    self._verify_operation(entity_name, operation)
                    return entity
            raise TypeError(f"Api entity {entity} is not supported.")

    def _verify_operation(self,
                          entity,
                          operation: str = None):
        """
        verify operation
        """
        if operation is not None and operation not in self._dto_registry[entity]["operations"]:
            raise NotImplementedError(f"Operation {operation} not supported for {entity} entity.")

    def get_components(self,
                       label_id: uuid.UUID = None,
                       twin_id: uuid.UUID = None,
                       template_id: uuid.UUID = None,
                       owner_id: uuid.UUID = None,
                       organization_only: bool = False,
                       name: str = None):
        """
        get components
        :param label_id: filter on a specific label
        :param template_id: filter on a specific template
        :param twin_id: filter on a specific twin
        :param owner_id: filter on a specific owner_id
        :param organization_only: work only with organization components (by default - False)
        :param name: filter on a specific name (contains)
        """
        pass

    def create_component(self, component: SolutionComponent):
        """
        create a component based on its ID.
        """
        pass

    def update_component(self, component: SolutionComponent):
        """
        update a component based on its ID.
        """
        pass

    def delete_component(self, component_id: uuid.UUID):
        """
        delete component
        """
        pass

    def get_business_labels(self) -> dict:
        """
        get a name / uuid dictionary with all business labels in platform.
        """
        pass

    def get_datapoint_mappings(self, registration):
        """
        get datapoint mapping from a registration.
        """
        pass

    def get_registrations(self, template) -> list:
        """
        retrieve all registrations for
        :param template: template object, UUID or str key.
        :return: list of twin registration.
        """
        pass

    def search_datapoints(self,
                          page: int = 1,
                          size: int = 20,
                          sort: str = "id",
                          direction: str = "asc",
                          hardware_id: str = None,
                          categories: list = None,
                          business_types: list = None,
                          twin=None,
                          recursive: bool = False) -> PagedQueryResult:
        """
        get datapoints with a paged query.
        :param page: numero of the page - default 1.
        :param size: quantity per page - default 20 max 100.
        :param sort: column to sort results - default id.
        :param direction: sorting direction by default asc, accept also desc.
        :param hardware_id: filter on a specific hardware ID name or partial name.
        :param business_types: list of BusinessType or str.
        :param categories: list of UUID or Category.
        :param twin: uuid or Twin element to search datapoints.
        :param recursive: set to True in combination of a twin to look inside all sub-twins recursively.
        :return: PagedQueryResults, check total for number of potential results and results for the list of entity.
        """
        pass

    def search_twins(self,
                     page: int = 1,
                     size: int = 20,
                     sort: str = "id",
                     direction: str = "asc",
                     hardware_id: str = None,
                     name: str = None,
                     parents: list = None) -> PagedQueryResult:
        """
        get twins with a paged query.
        :param page: numero of the page - default 1.
        :param size: quantity per page - default 20 max 100.
        :param sort: column to sort results - default id.
        :param direction: sorting direction by default asc, accept also desc.
        :param hardware_id: filter on a specific hardware ID name or partial name.
        :param name: name or part of twin name.
        :param parents: list of all possible parents (Twin, UUID, or str UUID).
        :return: PagedQueryResults, check total for number of potential results and results for the list of entity.
        """
        pass

    def search_insights(self,
                        page: int = 1,
                        size: int = 20,
                        sort: str = "id",
                        direction: str = "asc",
                        datapoint_id: str = None,
                        twin_id: str = None,
                        component_id: list = None):
        """
        get insights with a paged query.
        :param page: numero of the page - default 1.
        :param size: quantity per page - default 20 max 100.
        :param sort: column to sort results - default id.
        :param direction: sorting direction by default asc, accept also desc.
        :param datapoint_id: id of the datapoint on which filtering the insights.
        :param twin_id: id of the twin on which filtering the insights.
        :param component_id: id of the component on which filtering the insights.
        :return: PagedQueryResults, check total for number of potential results and results for the list of entity.
        """
        pass

    def send_alerts(self,
                    message: str,
                    recipients: list,
                    alert_type: AlertType = AlertType.SMS,
                    subject: str = "",
                    cc: list = None):
        """
        send alerts through backend
        :param str message: message to send.
        :param list recipients: list of recipients.
        :param wizata_dsapi.AlertType alert_type: type of alerts (default sms).
        :param str subject: for email only - subject of email.
        :param list cc: for email only - cc recipients list.
        """
        pass

    def download_model(self, identifier: str) -> ModelInfo:
        """
        download a model image directly from the repository, auto-select alias if necessary.
        :param identifier: exact identifier including alias of the model or no alias to auto-select active version.
        :return: ModelInfo with trained_model loaded, extra file must be downloaded separately.
        """
        pass

    def download_file(self,
                      model: ModelInfo = None,
                      file: ModelFile = None,
                      identifier: str = None,
                      path: str = None):
        """
        download a model extra file from the repository.
        :param model: model (ModelInfo), alternatively can use identifier
        :param file: file (ModelFile), alternatively can use path
        :param identifier: identifier including alias of the model.
        :param path: path name of the file including the extension.
        :return: bytes content.
        """
        pass

    def upload_file(self,
                    content,
                    model: ModelInfo = None,
                    file: ModelFile = None,
                    identifier: str = None,
                    path: str = None
                    ):
        """
        upload a model extra file to the repository.
        please use upload_model for a trained_model
        :param content: the bytes content of the file.
        :param model: model (ModelInfo), alternatively can use identifier
        :param file: file (ModelFile), alternatively can use path
        :param identifier: identifier including alias of the model.
        :param path: path name of the file including the extension.
        :return: bytes content.
        """
        pass

    def upload_model(self,
                     model_info: ModelInfo,
                     bytes_content=None):
        """
        upload a model within the model repository.
            - by default use model_info.trained_model and convert it to a pickle
            - for already torch or pickle please pass the bytes_content
            - model_info.file_format must be set properly to 'pkl' or 'pt'
        :param model_info: model info, with at least key (+twin, +property, +alias) and trained_model.
        :param bytes_content: bytes[] of your torch or pickle model.
        """
        pass


import uuid
from .api_dto import ApiDto
from enum import Enum
import re
import json


def is_valid_hex_color(color):
    hex_color_pattern = r'^#([A-Fa-f0-9]{6})$'
    match = re.match(hex_color_pattern, color)
    if match is None:
        raise ValueError('color is not a proper hexadecimal color : e.g. #E64600')
    else:
        return True


class TwinType(ApiDto):
    """
    twin type
    """

    @classmethod
    def route(cls):
        return "twintypes"

    @classmethod
    def from_dict(cls, data):
        obj = TwinType()
        obj.from_json(data)
        return obj

    def __init__(self,
                 twin_type_id: uuid.UUID = None,
                 name: str = None,
                 icon: str = None,
                 color: str = None,
                 description: str = None):
        if twin_type_id is None:
            self.twin_type_id = uuid.uuid4()
        else:
            self.twin_type_id = twin_type_id
        self.name = name
        self.icon = icon
        self.color = color
        if color is not None:
            is_valid_hex_color(color)
        self.description = description
        self.createdById = None
        self.createdDate = None
        self.updatedById = None
        self.updatedDate = None

    def api_id(self) -> str:
        return str(self.twin_type_id).upper()

    def endpoint(self) -> str:
        return "TwinTypes"

    def from_json(self, obj):
        if "id" in obj.keys():
            self.twin_type_id = uuid.UUID(obj["id"])
        if "name" in obj.keys() and obj["name"] is not None:
            self.name = obj["name"]
        if "color" in obj.keys() and obj["color"] is not None:
            is_valid_hex_color(str(obj["color"]))
            self.color = str(obj["color"])
        if "icon" in obj.keys() and obj["icon"] is not None:
            self.icon = str(obj["icon"])
        if "description" in obj.keys() and obj["description"] is not None:
            self.description = str(obj["description"])
        if "createdById" in obj.keys() and obj["createdById"] is not None:
            self.createdById = obj["createdById"]
        if "createdDate" in obj.keys() and obj["createdDate"] is not None:
            self.createdDate = obj["createdDate"]
        if "updatedById" in obj.keys() and obj["updatedById"] is not None:
            self.updatedById = obj["updatedById"]
        if "updatedDate" in obj.keys() and obj["updatedDate"] is not None:
            self.updatedDate = obj["updatedDate"]

    def to_json(self, target: str = None):
        obj = {
            "id": str(self.twin_type_id)
        }
        if self.name is not None:
            obj["name"] = str(self.name)
        if self.color is not None:
            obj["color"] = str(self.color)
        if self.description is not None:
            obj["description"] = str(self.description)
        if self.icon is not None:
            obj["icon"] = str(self.icon)
        if self.createdById is not None:
            obj["createdById"] = str(self.createdById)
        if self.createdDate is not None:
            obj["createdDate"] = str(self.createdDate)
        if self.updatedById is not None:
            obj["updatedById"] = str(self.updatedById)
        if self.updatedDate is not None:
            obj["updatedDate"] = str(self.updatedDate)
        return obj



class TwinBlockType(Enum):
    """
    deprecated - replaced by TwinType
    digital twin block type
        - area
        - machine
        - equipment
        - flow
    """
    AREA = "area"
    MACHINE = "machine"
    EQUIPMENT = "equipment"
    FLOW = "flow"


class Twin(ApiDto):
    """
    digital twin item representing an asset, machine, equipment.

    :ivar uuid.UUID twin_id: technical id of the twin.
    :ivar uuid.UUID hardware_id: unique str hardware id (e.g. required in template and ml models 'by_twin').
    :ivar str name: logical display name of the Asset.
    :ivar uuid.UUID parent_id: technical id of the parent twin item.
    :ivar wizata_dsapi.TwinBlockType type: type (area, machine, equipment or flow).
    :ivar float latitude: fixed latitude of the twin item.
    :ivar float longitude: fixed longitude of the twin item.
    :ivar uuid.UUID latitude_id: technical id of a datapoint specifying dynamic latitude of twin.
    :ivar uuid.UUID longitude_id: technical id of a datapoint specifying dynamic longitude of twin.
    :ivar str color: hexadecimal code of the color to associate with this twin item.
    :ivar str icon: logical id of the icon or svg schema to associate with this twin item.
    :ivar uuid.UUID createdById: unique identifier of creating user.
    :ivar int createdDate: timestamp of created date.
    :ivar uuid.UUID updatedById: unique identifier of updating user.
    :ivar int updatedDate: timestamp of updated date.
    """

    @classmethod
    def route(cls):
        return "twins"

    @classmethod
    def from_dict(cls, data):
        obj = Twin()
        obj.from_json(data)
        return obj

    def __init__(self,
                 twin_id=None,
                 hardware_id=None,
                 name=None, parent_id=None,
                 ttype=TwinBlockType.MACHINE,
                 type_id:uuid.UUID = None,
                 twin_type:TwinType = None,
                 latitude_id: uuid.UUID = None,
                 longitude_id: uuid.UUID = None,
                 color: str = None,
                 icon: str = None,
                 latitude: float = None,
                 longitude: float = None,
                 extra_properties: dict = None,
                 description: str = None):
        if twin_id is None:
            self.twin_id = uuid.uuid4()
        else:
            self.twin_id = twin_id
        self.hardware_id = hardware_id
        self.name = name
        self.parent_id = parent_id
        self.type = ttype
        if twin_type is not None:
            self.twin_type = twin_type
            self.type_id = twin_type.twin_type_id
        else:
            self.twin_type = None
            self.type_id = type_id
        self.description = description
        self.latitude = latitude
        self.longitude = longitude
        self.latitude_id = latitude_id
        self.longitude_id = longitude_id
        if color is not None:
            is_valid_hex_color(color)
        self.color = color
        self.icon = icon
        self.extra_properties = extra_properties
        self.createdById = None
        self.createdDate = None
        self.updatedById = None
        self.updatedDate = None

    def api_id(self) -> str:
        return str(self.twin_id).upper()

    def api_key(self) -> str:
        return self.hardware_id

    def set_id(self, id_value):
        self.twin_id = uuid.UUID(str(id_value))

    def endpoint(self) -> str:
        return "Twins"

    def from_json(self, obj):
        if "id" in obj.keys():
            self.twin_id = uuid.UUID(obj["id"])
        if "hardwareId" in obj.keys() and obj["hardwareId"] is not None:
            self.hardware_id = obj["hardwareId"]
        if "name" in obj.keys() and obj["name"] is not None:
            self.name = obj["name"]
        if "parentId" in obj.keys() and obj["parentId"] is not None:
            self.parent_id = uuid.UUID(obj["parentId"])
        if "twinTypeId" in obj.keys() and obj["twinTypeId"] is not None:
            self.type_id = uuid.UUID(obj["twinTypeId"])
        if "latitude" in obj.keys() and obj["latitude"] is not None:
            self.latitude = float(obj["latitude"])
        if "longitude" in obj.keys() and obj["longitude"] is not None:
            self.longitude = float(obj["longitude"])
        if "latitudeSensorId" in obj.keys() and obj["latitudeSensorId"] is not None:
            self.latitude_id = uuid.UUID(obj["latitudeSensorId"])
        if "longitudeSensorId" in obj.keys() and obj["longitudeSensorId"] is not None:
            self.longitude_id = uuid.UUID(obj["longitudeSensorId"])
        if "color" in obj.keys() and obj["color"] is not None:
            is_valid_hex_color(str(obj["color"]))
            self.color = str(obj["color"])
        if "icon" in obj.keys() and obj["icon"] is not None:
            self.icon = str(obj["icon"])
        if "description" in obj.keys() and obj["description"] is not None:
            self.description = str(obj["description"])
        if "type" in obj.keys():
            self.type = TwinBlockType(str(obj["type"]))
        if "extraProperties" in obj.keys():
            if isinstance(obj["extraProperties"], str):
                self.extra_properties = json.loads(obj["extraProperties"])
            else:
                self.extra_properties = obj["extraProperties"]
        if "createdById" in obj.keys() and obj["createdById"] is not None:
            self.createdById = obj["createdById"]
        if "createdDate" in obj.keys() and obj["createdDate"] is not None:
            self.createdDate = obj["createdDate"]
        if "updatedById" in obj.keys() and obj["updatedById"] is not None:
            self.updatedById = obj["updatedById"]
        if "updatedDate" in obj.keys() and obj["updatedDate"] is not None:
            self.updatedDate = obj["updatedDate"]

    def to_json(self, target: str = None):
        obj = {
            "id": str(self.twin_id)
        }
        if self.hardware_id is not None:
            obj["hardwareId"] = str(self.hardware_id)
        if self.name is not None:
            obj["name"] = str(self.name)
        if self.parent_id is not None:
            obj["parentId"] = str(self.parent_id)
        if self.twin_type is not None:
            obj["twinTypeId"] = str(self.twin_type.twin_type_id)
        elif self.type_id is not None:
            obj["twinTypeId"] = str(self.type_id)
        if self.type is not None and isinstance(self.type, TwinBlockType):
            obj["type"] = self.type.value
        if self.color is not None:
            obj["color"] = str(self.color)
        if self.description is not None:
            obj["description"] = str(self.description)
        if self.latitude is not None:
            obj["latitude"] = float(self.latitude)
        if self.longitude is not None:
            obj["longitude"] = float(self.longitude)
        if self.latitude_id is not None:
            obj["latitudeSensorId"] = str(self.latitude_id)
        if self.longitude_id is not None:
            obj["longitudeSensorId"] = str(self.longitude_id)
        if self.icon is not None:
            obj["icon"] = str(self.icon)
        if self.extra_properties is not None:
            if not isinstance(self.extra_properties, dict):
                raise ValueError('on twin extra properties must be None or a JSON serializable dict')
            obj["extraProperties"] = json.dumps(self.extra_properties)
        if self.createdById is not None:
            obj["createdById"] = str(self.createdById)
        if self.createdDate is not None:
            obj["createdDate"] = str(self.createdDate)
        if self.updatedById is not None:
            obj["updatedById"] = str(self.updatedById)
        if self.updatedDate is not None:
            obj["updatedDate"] = str(self.updatedDate)
        return obj



from dataclasses import dataclass, field
from typing import Dict, List, Any, Optional, Union


class FilterOperator:
    """
    Constants for standard filtering operators.
    """
    EQ = "eq"
    NE = "ne"
    GT = "gt"
    GTE = "gte"
    LT = "lt"
    LTE = "lte"
    IN = "in"
    NIN = "nin"
    CONTAINS = "contains"
    SEARCH = "_search"


class GroupSummaryType:
    """
    Aggregate methods for group summary types.
    """
    SUM = "Sum"
    AVG = "Avg"
    MAX = "Max"
    MIN = "Min"
    COUNT = "Count"

class GroupOption:
    def __init__(self, property_name: str):
        self.property_name = property_name

class GroupSummary:
    def __init__(self, type: str, property_name: str):
        self.type = type
        self.property_name = property_name


class SortOption:
    def __init__(self, field: str, order: str = "asc"):
        self.field = field
        self.order = order


class SearchQuery:
    """
    Generic class to represent JSON request structure.
    """
    @classmethod
    def from_dict(cls, data: dict):
        """
        Helper to convert dictionary coming from endpoint to SearchQuery object.
        """

        svc_name = data.get("serviceName") or data.get("service_name") or "pipeline_runner"

        return cls(
            filters=data.get("filters"),
            sort=data.get("sort"),
            limit=data.get("limit", 1000),
            page=data.get("page", 1),
            size=data.get("size", 50),
            service_name=svc_name
        )


    def __init__(self,
                 filters: Optional[Dict[str, Any]] = None,
                 sort: Optional[List[SortOption]] = None,
                 limit: int = 1000,
                 page: int = 1,
                 size: int = 50,
                 service_name: str = "pipeline_runner"):

        self.filters = filters if filters is not None else {}
        self.limit = limit
        self.page = page
        self.size = size
        self.service_name = service_name

        self.sort: List[SortOption] = []
        if sort:
            for s in sort:
                if isinstance(s, SortOption):
                    self.sort.append(s)
                elif isinstance(s, dict):
                    self.sort.append(SortOption(field=s.get("field"), order=s.get("order", "asc")))


    def to_dict(self) -> dict:
        """
        Convert SearchQuery to a dictionary matching the API expected JSON format.
        """
        data = {
            "filters": self.filters,
            "limit": self.limit,
            "page": self.page,
            "size": self.size,
            "serviceName": self.service_name,
        }
        if self.sort:
            data["sort"] = [{"field": s.field, "order": s.order} for s in self.sort]
        return data

class GroupSearchQuery:
    """
    Generic class to represent JSON request structure.
    """
    @classmethod
    def from_dict(cls, data: dict):
        """
        Helper to convert dictionary coming from endpoint to SearchQuery object.
        """

        svc_name = data.get("serviceName") or data.get("service_name") or "pipeline_runner"

        return cls(
            filters=data.get("filters"),
            service_name=svc_name,
            groups=data.get("groups"),
            group_summaries=data.get("groupSummaries")
        )


    def __init__(self,
                 filters: Optional[Dict[str, Any]] = None,
                 service_name: str = "pipeline_runner",
                 groups: Optional[List[Union[GroupOption, dict]]] = None,
                 group_summaries: Optional[List[Union[GroupSummary, dict]]] = None):

        self.filters = filters if filters is not None else {}
        self.service_name = service_name


        # New grouping feature
        self.groups: List[GroupOption] = []
        if groups:
            for g in groups:
                if isinstance(g, GroupOption):
                    self.groups.append(g)
                elif isinstance(g, dict) and g.get("propertyName"):
                    self.groups.append(GroupOption(property_name=g.get("propertyName")))


        # GroupSummaries
        self.group_summaries: List[GroupSummary] = []
        if group_summaries:
            for gs in group_summaries:
                if isinstance(gs, GroupSummary):
                    self.group_summaries.append(gs)
                elif isinstance(gs, dict) and gs.get("propertyName"):
                    sum_type = gs.get("type", "Count")
                    self.group_summaries.append(GroupSummary(type=sum_type, property_name=gs.get("propertyName")))




import json
import uuid
from .api_dto import ApiDto


class Evaluation(ApiDto):
    """
    an evaluation stores insights, such as properties and accuracy once a model training was executed.
    """

    @classmethod
    def route(cls):
        return "evaluations"

    @classmethod
    def from_dict(cls, data):
        obj = Evaluation()
        obj.from_json(data)
        return obj

    def __init__(self,
                 evaluation_id: uuid.UUID = None,
                 execution_id: int = None,
                 model_id: str = None,
                 properties: dict = None,
                 metrics: dict = None):
        if evaluation_id is None:
            self.evaluation_id = uuid.uuid4()
        else:
            self.evaluation_id = evaluation_id
        self.execution_id = execution_id
        self.model_id = model_id
        self.properties = properties
        self.metrics = metrics

        # related entities
        self.execution = None

    def api_id(self) -> str:
        return str(self.evaluation_id).upper()

    def endpoint(self) -> str:
        return "Evaluations"

    def to_json(self, target: str = None):
        obj = {
            "id": str(self.evaluation_id)
        }
        if self.execution_id is not None:
            obj["executionId"] = int(self.execution_id)
        if self.model_id is not None:
            obj["modelId"] = str(self.model_id)
        if self.properties is not None:
            obj["properties"] = json.dumps(self.properties)
        if self.metrics is not None:
            obj["metrics"] = json.dumps(self.metrics)
        return obj

    def from_json(self, obj):
        if "id" in obj.keys():
            self.evaluation_id = uuid.UUID(obj["id"])
        if "executionId" in obj.keys() and obj["executionId"] is not None:
            self.execution_id = int(obj["executionId"])
        if "modelId" in obj.keys() and obj["modelId"] is not None:
            self.model_id = obj["modelId"]
        if "properties" in obj.keys() and obj["properties"] is not None:
            self.properties = json.loads(obj["properties"])
        if "metrics" in obj.keys() and obj["metrics"] is not None:
            self.metrics = json.loads(obj["metrics"])



import sys
import uuid
from typing import Optional
from uuid import UUID

from .dataframe_toolkit import generate_epoch


def get_streamlit_token():
    if 'streamlit' in sys.modules:
        st = sys.modules['streamlit']
        if st.query_params is not None and "auth_token" in st.query_params:
            auth_token = st.query_params["auth_token"]
            return auth_token
    return None


def get_streamlit_domain():
    if 'streamlit' in sys.modules:
        st = sys.modules['streamlit']
        if st.query_params is not None and "dsapi" in st.query_params:
            domain = st.query_params["dsapi"]
            return domain
    return None

def get_streamlit_twin_id() -> Optional[UUID]:
    """
    return current selected twin_id.
    :return:
    """
    if 'streamlit' in sys.modules:
        st = sys.modules['streamlit']
        if st.query_params is not None and "twin_id" in st.query_params:
            twin_id = uuid.UUID(st.query_params["twin_id"])
            return twin_id
    return None


def get_streamlit_from() -> Optional[int]:
    """
    timestamp representation of from parameter on UI (ms)
    :return:
    """
    if 'streamlit' in sys.modules:
        st = sys.modules['streamlit']
        if st.query_params is not None and "from" in st.query_params:
            param = st.query_params["from"]
            if 'now' in param:
                return generate_epoch(param)
            else:
                return int(param)
    return None


def get_streamlit_to() -> Optional[int]:
    """
    timestamp representation of to parameter on UI (ms)
    :return:
    """
    if 'streamlit' in sys.modules:
        st = sys.modules['streamlit']
        if st.query_params is not None and "to" in st.query_params:
            param = st.query_params["to"]
            if 'now' in param:
                return generate_epoch(param)
            else:
                return int(param)
    return None



from .execution_log import ExecutionLog
import logging


class ILogger:
    """
    logger interface used within a pipeline and context.
    """

    def write_log(self, message: str = None, level: int = logging.INFO):
        """
        write a log
        :param str message: message to write.
        :param int level: use logging level.
        """
        pass

    def notify(self):
        """
        notify the listeners and watchers on current status.
        """
        pass



# Api Entities (Dto)
from .plot import Plot
from .datapoint import DataPoint
from .datastore import DataStore
from .mlmodel import ModelInfo, ModelList
from .request import Request
from .execution import Execution, ExecutionStatus
from .experiment import Experiment
from .evaluation import Evaluation
from .group_system import GroupSystem
from .script import Script
from .template import Template, TemplateProperty
from .solution_component import SolutionComponent
from .business_label import BusinessLabel
from .twinregistration import TwinRegistration, TwinRegistrationProperty
from .trigger import Trigger
from .twin import Twin, TwinType
from .pipeline import Pipeline
from .pipeline_image import PipelineImage
from .insight import Insight


# init APIs supported operations
_registry = {
    "business_labels":
        {
            "class": BusinessLabel,
            "cloud_dsapi": ["lists"],
            "cloud_context": []
        },
    "components":
        {
            "class": SolutionComponent,
            "cloud_dsapi": ['lists', 'get_by_id', 'create', 'update', 'delete'],
            "cloud_context": ['get_by_id', 'create', 'update', 'delete']
        },
    "datapoints":
        {
            "class": DataPoint,
            "cloud_dsapi": ['lists', 'get_by_id', 'get_by_key', 'create', 'update', 'delete', 'search'],
            "cloud_context": ['get_by_id', 'get_by_key', 'create', 'update', 'delete', 'search']
        },
    "datastores":
        {
            "class": DataStore,
            "cloud_dsapi": ['lists', 'get_by_id', 'create', 'update', 'delete'],
            "cloud_context": []
        },
    "evaluations":
        {
            "class": Evaluation,
            "cloud_dsapi": ['lists', 'get_by_id', 'create', 'update', 'delete'],
            "cloud_context": ['get_by_id', 'create', 'update', 'delete']
        },
    "executions":
        {
            "class": Execution,
            "cloud_dsapi": ['get_by_id', 'search', 'abort'],
            "cloud_context": []
        },
    "experiments":
        {
            "class": Experiment,
            "cloud_dsapi": ['lists', 'get_by_id', 'get_by_key', 'create', 'update', 'delete'],
            "cloud_context": []
        },
    "groupsystems":
        {
            "class": GroupSystem,
            "cloud_dsapi": ['lists', 'get_by_id', 'get_by_key', 'create', 'update', 'delete'],
            "cloud_context": []
        },
    "insights":
        {
            "class": Insight,
            "cloud_dsapi": ['lists', 'get_by_id', 'create', 'update', 'delete', 'search'],
            "cloud_context": ['get_by_id', 'create', 'update', 'delete', 'search']
        },
    "models":
        {
            "class": ModelInfo,
            "cloud_dsapi": ['download_model', 'download_file' , 'upload_model'  , 'upload_file'],
            "cloud_context": ['download_model' , 'download_file' , 'upload_model'  , 'upload_file']
        },
    "pipelines":
        {
            "class": Pipeline,
            "cloud_dsapi": ['lists', 'get_by_id', 'get_by_key', 'create', 'update', 'delete'],
            "cloud_context": []
        },
    "pipelineimages":
        {
            "class": PipelineImage,
            "cloud_dsapi": ['lists', 'get_by_id', 'delete', 'build_image', 'download_image'],
            "cloud_context": []
        },
    "plots":
        {
            "class": Plot,
            "cloud_dsapi": ['lists', 'get_by_id', 'get_by_key', 'delete'],
            "cloud_context": []
        },
    "registrations":
        {
            "class": TwinRegistration,
            "cloud_dsapi": ['lists', 'get_by_id', 'create', 'update', 'delete'],
            "cloud_context": ['get_by_id', 'create', 'update', 'delete']
        },
    "registrationproperties":
        {
            "class": TwinRegistrationProperty,
            "cloud_dsapi": ['lists', 'get_by_id', 'create', 'update', 'delete'],
            "cloud_context": []
        },
    "request":
        {
            "class": Request,
            "cloud_dsapi": ['query'],
            "cloud_context": ['query']
        },
    "scripts":
        {
            "class": Script,
            "cloud_dsapi": ['lists', 'get_by_id', 'get_by_key', 'create', 'update', 'delete'],
            "cloud_context": []
        },
    "templates":
        {
            "class": Template,
            "cloud_dsapi": ['lists', 'get_by_id', 'get_by_key', 'create', 'update', 'delete'],
            "cloud_context": ['get_by_id', 'get_by_key', 'create', 'update', 'delete']
        },
    "templateproperties":
        {
            "class": TemplateProperty,
            "cloud_dsapi": ['lists', 'get_by_id', 'create', 'update', 'delete'],
            "cloud_context": []
        },
    "triggers":
        {
            "class": Trigger,
            "cloud_dsapi": ['lists', 'get_by_id', 'create', 'update', 'delete'],
            "cloud_context": ['get_by_id']
        },
    "twins":
        {
            "class": Twin,
            "cloud_dsapi": ['lists', 'get_by_id', 'get_by_key', 'create', 'update', 'delete', 'search'],
            "cloud_context": ['get_by_id', 'get_by_key', 'create', 'update', 'delete', 'search']
        },
    "twintypes":
        {
            "class": TwinType,
            "cloud_dsapi": [ 'get_by_id', 'create', 'update', 'delete'],
            "cloud_context": [ 'get_by_id', 'create', 'update', 'delete' ]
        }
}



import datetime
from datetime import timedelta
import json
import time
import uuid
import sys
from typing import Tuple

import dill
import requests
import pickle
import pandas
import plotly
import msal
import os
import types

import wizata_dsapi
import urllib.parse
import base64
import joblib
import io

import string
import random

from .api_dto import ApiDto, VarType, ApiDtoInterface
from .business_label import BusinessLabel
from .plot import Plot
from .request import Request
from .mlmodel import ModelInfo, ModelList, ModelFile
from .experiment import Experiment
from .script import Script
from .execution import Execution, ExecutionStatus
from .execution_log import ExecutionLog
from .search import SearchQuery
from .dsapi_json_encoder import DSAPIEncoder
from .group_system import GroupSystem
from .ds_dataframe import DSDataFrame
from .insight import Insight
from .model_toolkit import predict
from .template import Template, TemplateProperty
from .twinregistration import TwinRegistration
from .trigger import Trigger
from .twin import Twin
from .datapoint import DataPoint, BusinessType, Label, Category
from .pipeline import Pipeline, PipelineStep
from .pipeline_image import PipelineImage
from .solution_component import SolutionComponent, SolutionType
from .paged_query_result import PagedQueryResult
from .api_interface import ApiInterface
from .version import __version__
import ast
from .api_config import _registry

from .streamlit_utils import get_streamlit_token, get_streamlit_domain


def sanitize_url_parameter(param_value):
    illegal_characters = set(c for c in param_value if not (c.isalnum() or c in '-_., '))
    if illegal_characters:
        raise ValueError(f"illegal characters found in parameter {param_value}: {', '.join(illegal_characters)}")
    encoded_param_value = urllib.parse.quote(param_value)
    return encoded_param_value


def parse_string_list(s):
    try:
        result = ast.literal_eval(s)
        if isinstance(result, list):
            return result
        else:
            return []
    except (ValueError, SyntaxError):
        return []


class WizataDSAPIClient(ApiInterface, ApiDtoInterface):
    """
    client wrapper to cloud data science API

    accessible preferably through wizata_dsapi.api() using os variables, you can also create an instance using __init__()
    please contact your administrator to receive configuration and connection details

    :ivar str domain: URL of data science API
    :ivar str client_id: azure authentication - client id
    :ivar str client_secret: azure authentication - your client secret API key
    :ivar str scope: azure authentication - scope
    :ivar str tenant_id: azure authentication - tenant
    :ivar str protocol: default - https
    :ivar str username: azure authentication - deprecated - username
    :ivar str password: azure authentication - deprecated - password
    """

    def __init__(self,
                 client_id=None,
                 scope=None,
                 tenant_id=None,
                 username=None,
                 password=None,
                 domain="localhost",
                 protocol="https",
                 client_secret=None):
        super().__init__()
        super()._set_registry("cloud_dsapi", _registry)

        # properties
        self.domain = domain
        self.protocol = protocol

        # authentication
        self.__username = username
        self.__password = password

        self.__client_id = client_id
        self.__tenant_id = tenant_id
        if tenant_id is not None:
            self.__authority = "https://login.microsoftonline.com/" + tenant_id
        self.__scopes = [scope]

        self.__interactive_token = None
        self.__daemon = False

        if client_secret is not None:
            self.__daemon = True
            self.__confidential_app = msal.ConfidentialClientApplication(
                client_id=self.__client_id,
                client_credential=client_secret,
                authority=self.__authority
            )

        if self.__client_id is not None:
            self.__app = msal.PublicClientApplication(
                client_id=self.__client_id,
                authority=self.__authority
            )

        self.roles = []
        self.twins = []

    def authenticate(self):
        """
        perform authentication on client side
        """
        result = self.__app.acquire_token_interactive(
            scopes=self.__scopes,
            success_template="""<html><body>You are authenticated and your code is running, you can close this page.<script>setTimeout(function(){window.close()}, 3000);</script></body></html> """
        )
        self.__daemon = False
        self.__interactive_token = result["access_token"]

    def __url(self):
        return self.protocol + "://" + self.domain + "/dsapi/"

    def __token(self):
        # Interactive Authentication
        if self.__interactive_token is not None:
            return self.__interactive_token

        streamlit_token = get_streamlit_token()
        if streamlit_token is not None:
            return streamlit_token

        if not self.__daemon:
            # Silent Authentication
            result = None
            accounts = self.__app.get_accounts(username=self.__username)
            if accounts:
                # If there is an account in the cache, try to get the token silently
                result = self.__app.acquire_token_silent(scopes=self.__scopes, account=accounts[0])

            if not result:
                # If there is no cached token, try to get a new token using the provided username and password
                result = self.__app.acquire_token_by_username_password(
                    username=self.__username,
                    password=self.__password,
                    scopes=self.__scopes
                )

            if "error_description" in result.keys():
                raise RuntimeError(str(result["error_description"]))

            if "access_token" in result.keys():
                return result["access_token"]
            else:
                raise RuntimeError(str(result))
        else:
            result = self.__confidential_app.acquire_token_silent(scopes=self.__scopes, account=None)

            if not result:
                result = self.__confidential_app.acquire_token_for_client(scopes=self.__scopes)

            if "error_description" in result.keys():
                raise RuntimeError(str(result["error_description"]))

            if "access_token" in result.keys():
                return result["access_token"]

            return result

    def __header(self):
        return {
            'Content-Type': 'application/json',
            'Authorization': f'Bearer {self.__token()}',
            'Wizata-Version': f'{__version__}'
        }

    def __raise_error(self, response):
        """
        raise a formatted error from an API response.
        :param response: response from the server.
        """
        if 'Content-Type' in response.headers and response.headers['Content-Type'] == 'application/json':
            json_content = response.json()
            if "errors" in json_content.keys():
                message = json_content["errors"][0]["message"]
                return RuntimeError(str(response.status_code) + " - " + message)
            else:
                return RuntimeError(str(response.status_code) + " - " + response.reason)
        else:
            raise RuntimeError(str(response.status_code) + " - " + response.reason)

    def access_check(self):
        """
        call the who am I request, to read proper headers and get specific information.
        :return: roles, twins
        """
        self._process_request(method='GET', route='whoami/')
        return self.roles, self.twins

    def who_am_i(self):
        """
        decrypt current JWT token used to get information about current user.
        :return:
        """
        token = self.__token()
        if token is None:
            return {}
        parts = token.split('.')
        if len(parts) != 3:
            return None  # not a JWT
        _, payload_b64, _ = parts
        def _b64url_decode(input_str: str) -> bytes:
            rem = len(input_str) % 4
            if rem:
                input_str += '=' * (4 - rem)
            return base64.urlsafe_b64decode(input_str.encode('utf-8'))
        payload_json = json.loads(_b64url_decode(payload_b64).decode('utf-8'))
        return payload_json

    def _process_request(self,
                         method: str,
                         route: str,
                         dto_class=None,
                         result_type: str = None,
                         params: dict = None,
                         data=None):
        """
        process a request against DS API.
        :param method: HTTP operation method to use ( GET, POST, DELETE, PUT, ... )
        :param route: sub-route from main URL to target.
        :param dto_class: set a ApiDto class implementation to use (e.g. to format result)
        :param params: query parameters to add if presents
        :param result_type: set if result should be processed ('list', 'object')
        :param data: data for POST, PUT, PATCH requests
        """
        response = None
        if method == "GET":
            response = requests.request(
                "GET",
                self.__url() + route,
                headers=self.__header(),
                params=params
            )
        if method == "POST":
            response = requests.post(
                self.__url() + route,
                headers=self.__header(),
                params=params,
                data=data
            )
        if method == "PUT":
            response = requests.put(
                self.__url() + route,
                headers=self.__header(),
                params=params,
                data=data
            )
        if method == "DELETE":
            response = requests.delete(
                self.__url() + route,
                headers=self.__header(),
                params=params
            )
        if response is None:
            raise ValueError('Method not supported')
        if "Wizata-Version" in response.headers:
            server_version = response.headers["Wizata-Version"]
            if server_version != wizata_dsapi.__version__:
                print(f' * Warning - version mismatch - server {server_version} vs local {wizata_dsapi.__version__}')
                print(f'        python -m pip install --upgrade pip')
                print(f'        pip install wizata_dsapi=={server_version}')
        if response.status_code == 200:
            if 'Wizata-Roles' in response.headers:
                self.roles = parse_string_list(response.headers['Wizata-Roles'])
            if 'Wizata-Twins' in response.headers:
                self.twins = parse_string_list(response.headers['Wizata-Twins'])
            if result_type == "list":
                obj_list = []
                for obj_json in response.json():
                    obj_list.append(dto_class.from_dict(obj_json))
                return obj_list
            elif result_type == "json":
                return dto_class.from_dict(response.json())
            elif result_type == "pickle":
                pickle_bytes = response.content
                return pickle.loads(pickle_bytes)
            elif result_type == "dill":
                script_bytes = response.content
                return dill.loads(script_bytes)
            elif result_type == "bytes":
                content_bytes = response.content
                return content_bytes
            else:
                return response
        else:
            raise self.__raise_error(response)

    def _query(self, request: wizata_dsapi.Request) -> pandas.DataFrame:
        """
        Perform a request on Time-Series through DS API
        """
        return self._process_request(
            method="POST",
            route="execute/data",
            result_type="pickle",
            data=json.dumps(request.to_json(), cls=DSAPIEncoder)
        )

    def info(self):
        """
        print insights regarding version and supported operations
        """
        print(f"")
        print(f"version: {__version__} ")
        for class_name in _registry:
            print(f'Entity {class_name}')
            print(f' * Class    {_registry[class_name]["class"]} - ')
            print(f' * DS API   {_registry[class_name]["cloud_dsapi"]} - ')
            print(f' * Context  {_registry[class_name]["cloud_context"]}')
            print(f"")
        print(f"")

    def lists(self, entity):
        """
        lists all elements of a specific entity.
        :param entity: plural name of the entity or class (e.g. scripts, plots, mlmodels, dataframes...)
        :return: list of all elements with at least the id property.
        """
        dto_class = self._get_class(entity=entity, operation="lists")
        return self._process_request(
            method="GET",
            route=dto_class.route() + "/",
            dto_class=dto_class,
            result_type="list"
        )

    def get(self,
            obj=None,
            id: uuid.UUID = None,
            key: str = None,
            entity=None,
            script_name=None,
            experiment_key=None,
            pipeline_key=None,
            model_key=None,
            template_key=None,
            twin_hardware_id=None,
            datapoint_hardware_id=None):
        """
        get record content from DS API.

        - get will look first for obj then for id then for key (e.g. if id and key specified, key is ignored )

        :param obj: a wizata_dsapi.<Entity>() with UUID set - fetch using technical UUID

        :param id: a UUID of a specific object (use in combination of entity)
        :param key: a logical key of a specific object (e.g. hardware id for DataPoint and Twin (use in combination of entity)
        :param entity: plural name of the entity or class (use in combination of entity)(e.g. scripts, plots, mlmodels, dataframes...)
        :return: object with all properties or None if not found.

        """
        operation = "get_by_id"
        if obj is not None:
            if isinstance(obj, Request):
                return self._query(request=obj)
            else:
                dto_class = self._get_class(entity=type(obj), operation=operation)
                if dto_class == PipelineImage:
                    id = obj.api_id()
                else:
                    id = uuid.UUID(obj.api_id())
        elif entity is not None:
            if id is not None:
                dto_class = self._get_class(entity=entity, operation=operation)
            elif key is not None:
                operation = "get_by_key"
                dto_class = self._get_class(entity=entity, operation=operation)
            else:
                raise KeyError(f"cannot get record - get have an entity specified but no id or key")

        # deprecated exceptions
        elif script_name is not None:
            operation = "get_by_key"
            key = script_name
            dto_class = self._get_class(entity="scripts", operation=operation)
        elif experiment_key is not None:
            operation = "get_by_key"
            key = experiment_key
            dto_class = self._get_class(entity="experiments", operation=operation)
        elif model_key is not None:
            operation = "get_by_key"
            key = model_key
            dto_class = self._get_class(entity="mlmodels", operation=operation)
        elif pipeline_key is not None:
            operation = "get_by_key"
            key = pipeline_key
            dto_class = self._get_class(entity="pipelines", operation=operation)
        elif template_key is not None:
            operation = "get_by_key"
            key = template_key
            dto_class = self._get_class(entity="templates", operation=operation)
        elif twin_hardware_id is not None:
            operation = "get_by_key"
            key = twin_hardware_id
            dto_class = self._get_class(entity="twins", operation=operation)
        elif datapoint_hardware_id is not None:
            operation = "get_by_key"
            key = datapoint_hardware_id
            dto_class = self._get_class(entity="datapoints", operation=operation)

        # unsupported parameter combination
        else:
            raise KeyError(f"cannot get record - misconfiguration please provide valid obj or id+entity, key+entity")

        # get_by_id
        if operation == "get_by_id":
            return self._process_request(
                method="GET",
                route=dto_class.route() + "/" + str(id) + "/",
                dto_class=dto_class,
                result_type=dto_class.get_type()
            )

        # get_by_key
        elif operation == "get_by_key":
            if dto_class == Script:
                return self._process_request(
                    method="GET",
                    route=dto_class.route() + "/" + str(key) + "/",
                    dto_class=dto_class,
                    result_type=dto_class.get_type()
                )
            else:
                results = self._process_request(
                        method="GET",
                        route=dto_class.route() + "/",
                        dto_class=dto_class,
                        result_type="list",
                        params={'key': key}
                    )
                for result in results:
                    return result
                return None

    def create(self, obj) -> uuid.UUID:
        """
        create and save an object on the server
        :param obj: object from any supported entity (see info()) or python callable function (Script)
        :return: id of created object
        """
        if callable(obj) and isinstance(obj, types.FunctionType):
            obj = Script(
                function=obj
            )
        if not callable(obj) and isinstance(obj, Script):
            dto_class = self._get_class(entity=type(obj), operation="create")
            script = self._process_request(
                method="POST",
                route=dto_class.route() + "/",
                dto_class=dto_class,
                data=dill.dumps(obj),
                result_type="json")
            return script.script_id
        else:
            dto_class = self._get_class(entity=type(obj), operation="create")
            response = self._process_request(
                method="POST",
                route=dto_class.route() + "/",
                dto_class=dto_class,
                data=json.dumps(obj.to_json())
            )
            if dto_class.get_id_type() == int:
                obj.set_id(response.json()["id"])
                return response.json()["id"]
            else:
                return uuid.UUID(obj.api_id())

    def update(self, obj):
        """
        update and save an object on DS API
        :param obj: object to update on DS API (see info()) or python callable function (Script)
        """
        if callable(obj) and isinstance(obj, types.FunctionType):
            obj = Script(
                function=obj
            )
        if not callable(obj) and isinstance(obj, Script):
            dto_class = self._get_class(entity=type(obj), operation="update")
            self._process_request(
                method="PUT",
                route=dto_class.route() + "/" + str(obj.script_id) + "/",
                dto_class=dto_class,
                data=dill.dumps(obj))
        else:
            dto_class = self._get_class(entity=type(obj), operation="update")
            if dto_class.get_id_type() == uuid.UUID:
                id = uuid.UUID(obj.api_id())
            else:
                id = obj.api_id()
            self._process_request(
                method="PUT",
                route=dto_class.route() + "/" + str(id) + "/",
                dto_class=dto_class,
                data=json.dumps(obj.to_json())
            )

    def delete(self, obj):
        """
        delete an object on the server
        :param obj: object to delete including all content
        """
        dto_class = self._get_class(entity=type(obj), operation="delete")
        self._process_request(
            method="DELETE",
            route=dto_class.route() + "/" + obj.api_id() + "/"
        )

    def get_categories(self) -> dict:
        """
        get a name / uuid dictionary with all categories in platform.
        """
        response = requests.request("GET",
                                    self.__url() + "datapoints/categories/",
                                    headers=self.__header()
                                    )
        if response.status_code == 200:
            categories_json = response.json()
            categories_dict = {}
            for category_json in categories_json:
                categories_dict[category_json["name"]] = category_json["id"]
            return categories_dict
        else:
            raise self.__raise_error(response)

    def get_units(self) -> dict:
        """
        get a name / uuid dictionary with all units in platform.
        """
        response = requests.request("GET",
                                    self.__url() + "datapoints/units/",
                                    headers=self.__header()
                                    )
        if response.status_code == 200:
            units_json = response.json()
            units_dict = {}
            for unit in units_json:
                units_dict[unit["shortName"]] = unit["id"]
            return units_dict
        else:
            raise self.__raise_error(response)

    def get_business_labels(self) -> dict:
        """
        get a name / uuid dictionary with all business labels in platform.
        """
        response = requests.request("GET",
                                    self.__url() + "components/labels/",
                                    headers=self.__header()
                                    )
        if response.status_code == 200:
            categories_json = response.json()
            categories_dict = {}
            for category_json in categories_json:
                categories_dict[category_json["name"]] = category_json["id"]
            return categories_dict
        else:
            raise self.__raise_error(response)

    def upsert(self, obj):
        """
        upsert an object on the server using its logical unique key (e.g. hardware_id, key, name).
        if the object already exists (matched by key), it is updated. otherwise, it is created.

        :param obj: object to upsert on the server (must implement api_key()) or python callable function (Script)
        :return: ID of the object created or updated
        """
        if callable(obj) and isinstance(obj, types.FunctionType):
            obj = Script(
                function=obj
            )
        if isinstance(obj, Script):
            function = dill.dumps(obj.function)
            response = requests.post(self.__url() + f"scripts/{obj.name}",
                                    headers=self.__header(),
                                    data=function)
            if response.status_code == 200:
                obj.script_id = uuid.UUID(response.json()["id"])
                return obj.script_id
            else:
                raise self.__raise_error(response)

        key = obj.api_key()
        if key is None:
            raise TypeError(f"Type {type(obj).__name__} does not support upsert (no logical key).")

        existing = self.get(key=key, entity=type(obj))
        if existing is not None:
            obj.set_id(existing.api_id())
            return self.update(obj)
        else:
            return self.create(obj)

    def query(self,
              datapoints: list = None,
              start: datetime = None,
              end: datetime = None,
              interval: int = None,
              agg_method: str = "mean",
              template: str = None,
              twin: str = None,
              null: str = None,
              filters: dict = None,
              options: dict = None,
              group = None,
              field=None,
              bucket: str = None,
              tags: dict = None) -> pandas.DataFrame:
        """
        Query a dataframe from API.
        :param agg_method:
        :param datapoints: list of datapoints to fetch.
        :param start: start datetime of range to fetch
        :param end: end datetime of range to fetch
        :param interval: interval in milliseconds.
        :param template: template to fetch.
        :param twin: hardware ID of twin to fetch based on template.
        :param null: By default at 'drop' and dropping NaN values. If not intended behavior please set it to 'ignore' or 'all'.
        :param group: can be use to set group system and event retrieving instructions (dict or list[dict]).
        :param filters: dict of filters.
        :param options: dict of options.
        :param field: by default 'value' if none, accept str or list (e.g. value, reliability, eventId, ...)
        :param bucket: specify on which bucket the query applies.
        :return: dataframe
        """
        request = Request()

        if datapoints is not None:
            request.datapoints = datapoints

        if start is not None:
            request.start = start
        elif group is None:
            raise ValueError('start datetime is required.')

        if end is not None:
            request.end = end
        elif group is None:
            raise ValueError('end datetime is required.')

        if null is not None:
            request.null = null

        request.set_aggregation(agg_method, interval)

        if twin is not None:
            request.set_twin(twin)

        if template is not None:
            request.set_template(template)

        if filters is not None:
            request.filters = filters

        if tags is not None:
            request.tags = tags

        if options is not None:
            request.options = options

        if group is not None:
            request.group = group

        if field is not None:
            if not isinstance(field, str) and not isinstance(field, list):
                raise TypeError('field must be str or list')
            request.field = field

        if bucket is not None:
            request.bucket = bucket

        return self._query(request=request)

    def batch_query(self,
                    step: timedelta,
                    datapoints: list = None,
                    start: datetime = None,
                    end: datetime = None,
                    interval: int = None,
                    agg_method: str = "mean",
                    template: str = None,
                    twin: str = None,
                    null: str = None,
                    filters: dict = None,
                    options: dict = None,
                    group: dict = None,
                    field=None,
                    bucket: str = None,
                    tags: dict = None,
                    verbose: bool = True) -> pandas.DataFrame:
        all_data = []
        current_start = start
        total_duration = (end - start).total_seconds()

        while current_start < end:
            current_end = min(current_start + step, end)

            progress = ((current_start - start).total_seconds() / total_duration) * 100
            if verbose:
                print(f"Querying from {current_start} to {current_end}... ({progress:.2f}% completed)")

            df_batch = wizata_dsapi.api().query(
                datapoints=datapoints,
                start=current_start,
                end=current_end,
                interval=interval,
                agg_method=agg_method,
                template=template,
                twin=twin,
                null=null,
                filters=filters,
                options=options,
                group=group,
                field=field,
                bucket=bucket,
                tags=tags
            )

            all_data.append(df_batch)
            current_start = current_end

        if verbose:
            print("100% completed. Concatenating results...")

        df_final = pandas.concat(all_data)

        return df_final

    def get_ts_query(self,
                     datapoints: list[str] = None,
                     start: datetime = None,
                     end: datetime = None,
                     interval: int = None,
                     agg_method: str = "mean",
                     template: str = None,
                     twin: str = None,
                     null: str = None,
                     filters: dict = None,
                     options: dict = None) -> str:
        """
        Get a Query string to Timeseries Database.
        :param agg_method:
        :param datapoints: list of datapoints to fetch.
        :param start: start datetime of range to fetch
        :param end: end datetime of range to fetch
        :param interval: interval in milliseconds.
        :param template: template to fetch.
        :param twin: hardware ID of twin to fetch based on template.
        :param null: By default at 'drop' and dropping NaN values. If not intended behavior please set it to 'ignore' or 'all'.
        :param filters: dict of filters.
        :param options: dict of options.
        :return: dataframe
        """
        request = Request()

        if datapoints is not None:
            request.add_datapoints(datapoints)

        if start is not None:
            request.start = start
        else:
            raise ValueError('start datetime is required.')

        if end is not None:
            request.end = end
        else:
            raise ValueError('end datetime is required.')

        if null is not None:
            request.null = null

        if interval is None:
            raise ValueError('interval is required.')
        request.set_aggregation(agg_method, interval)

        if template is not None and twin is not None:
            request.select_template(
                template_key=template,
                twin_hardware_id=twin
            )

        if filters is not None:
            request.filters = filters

        if options is not None:
            request.options = options

        response = requests.request("POST", self.__url() + "data/string",
                                    headers=self.__header(),
                                    data=json.dumps(request.to_json(), cls=DSAPIEncoder))
        if response.status_code == 200:
            return response.json()
        else:
            raise self.__raise_error(response)

    def __execute(self,
                  execution_log: ExecutionLog = None,
                  experiment=None,
                  properties: dict = None,
                  pipeline=None,
                  image: str = None,
                  twin=None,
                  mode: str = 'experiment',
                  version: str = None,
                  ) -> ExecutionLog:
        """
        wrapper to execute a pipeline, used by other functions.
        """
        # Prepare
        if not execution_log:
            execution_log = ExecutionLog(
                experiment=experiment,
                properties=properties,
                pipeline=pipeline,
                twin=twin,
                pipeline_image_id=image,
                version=version
            )

        if isinstance(execution_log, ExecutionLog):
            dict_tmp = execution_log.to_dict()
            response = requests.post(f"{self.__url()}execute/?mode={mode}",
                                     headers=self.__header(),
                                     data=json.dumps(dict_tmp, cls=DSAPIEncoder))

            # Parse
            if response.status_code == 200:
                obj = response.json()
                return ExecutionLog.from_dict(obj)
            else:
                raise self.__raise_error(response)
        else:
            raise TypeError("No execution have been loaded from parameters.")

    def experiment(self,
                   pipeline,
                   experiment=None,
                   twin=None,
                   image: str = None,
                   properties: dict = None,
                   train: bool = True,
                   plot: bool = True,
                   write: bool = False,
                   version: str = None) -> ExecutionLog:
        """
        experiment and train models with a pipeline.

        - existing experiment is required (use create or upsert_experiment(key, name)).
        - if your pipeline is templated please provide a twin.
        - please provide all variables and parameters required through properties.
        - return an execution
            - check status with "wizata_dsapi.api().get(execution).status"
            - see plots with "wizata_dsapi.api().plots(execution)"

        :param experiment: existing experiment identified by key (str).
        :param pipeline: pipeline identified by key (str) .
        :param twin: twin identified by hardware ID (str)(optional).
        :param image: pipeline image id to use.
        :param properties: dictionary containing override for variables or additional parameters for your script.
        :param train: train machine learning model on model steps.
        :param plot: if False plot steps are ignored.
        :param write: if False write steps are ignored.
        :param version: version to use, by default use your python environment version.
        """
        if not train or not plot or write:
            if properties is None:
                properties = {}
            if "execution_options" not in properties:
                properties["execution_options"] = {}
            properties["execution_options"]["train"] = train
            properties["execution_options"]["plot"] = plot
            properties["execution_options"]["write"] = write

        return self.__execute(
            experiment=experiment,
            pipeline=pipeline,
            twin=twin,
            image=image,
            properties=properties,
            mode='experiment',
            version=version
        )

    def run(self,
            pipeline, twin=None,
            properties: dict = None,
            image: str = None,
            train: bool = False,
            plot: bool = False,
            write: bool = True,
            version: str = None) -> ExecutionLog:
        """
        run a pipeline.

        - existing models are used for simulation and prediction.
        - caution this might affect data inside platform or trigger automation.
        - if your pipeline is templated please provide a twin.
        - please provide all variables and parameters required through properties.
        - return an execution
            - check status with "wizata_dsapi.api().get(execution).status"
            - check results in platform (dashboard/explorer) or perform queries.

        :param pipeline: pipeline identified by its id (uuid or wizata_dsapi.Pipeline) or key (str).
        :param twin: twin identified by its id (uuid or wizata_dsapi.Twin) or hardware ID (str).
        :param properties: dictionary containing override for variables or additional parameters for your script.
        :param train: train machine learning model on model steps.
        :param image: pipeline image id to use.
        :param plot: if False plot steps are ignored.
        :param write: if False write steps are ignored.
        :param version: version of python environment to use if different than local environment.
        """

        if train or plot or not write:
            if properties is None:
                properties = {}
            if "execution_options" not in properties:
                properties["execution_options"] = {}
            properties["execution_options"]["train"] = train
            properties["execution_options"]["plot"] = plot
            properties["execution_options"]["write"] = write

        return self.__execute(
            pipeline=pipeline,
            twin=twin,
            properties=properties,
            image=image,
            mode='production',
            version=version
        )

    def await_execution(self,
                        execution: ExecutionLog,
                        timeout: int = 180,
                        poll_interval: int = 5) -> ExecutionLog:
        """
        wait for an execution to reach a terminal status (completed, failed or aborted).

        :param execution: ExecutionLog returned by experiment() or run().
        :param timeout: maximum seconds to wait before raising TimeoutError (default 180).
        :param poll_interval: seconds between status checks (default 5).
        :return: ExecutionLog with terminal status.
        :raises TimeoutError: if execution does not complete within timeout.
        :raises RuntimeError: if execution finished with failed or aborted status.
        """
        if not isinstance(execution, ExecutionLog):
            raise TypeError(f"expected ExecutionLog, got {type(execution)}")

        execution_id = execution.execution_id
        service_name = "experiment_runner" if execution.experiment is not None else "pipeline_runner"

        terminal = {ExecutionStatus.COMPLETED, ExecutionStatus.FAILED, ExecutionStatus.ABORTED}
        elapsed = 0

        while elapsed < timeout:
            time.sleep(poll_interval)
            elapsed += poll_interval

            result = self.search_execution({
                "filters": {"id": {"eq": str(execution_id)}},
                "serviceName": service_name,
            })
            for log in result.results:
                if log.status in terminal:
                    if log.status != ExecutionStatus.COMPLETED:
                        raise RuntimeError(
                            f"execution {execution_id} finished with status '{log.status.value}'. "
                            f"Messages: {log.messages}"
                        )
                    return log

        raise TimeoutError(
            f"execution {execution_id} did not complete within {timeout}s"
        )

    def plot(self, plot_id: str = None, plot: Plot = None, figure=None):
        """
        Fetch and show plot.
        :param plot: Wizata Plot Object
        :param figure: JSON Figure
        :param plot_id: Plot Id
        :return: plotly figure
        """
        if plot is not None and plot.figure is not None:
            return plotly.io.from_json(plot.figure)
        elif plot is not None and plot.figure is None:
            plot = self.get(plot)
            if plot.figure is not None:
                return plotly.io.from_json(plot.figure)
            else:
                raise ValueError('No plot has been fetch.')
        elif figure is not None:
            return plotly.io.from_json(plot.figure)
        elif plot_id is not None:
            plot = self.get(Plot(plot_id=plot_id))
            if plot.figure is not None:
                return plotly.io.from_json(plot.figure)
            else:
                raise ValueError('No plot has been fetch.')
        else:
            raise KeyError('No valid arguments.')

    def plots(self, execution: ExecutionLog):
        """
        get all plot for an execution.
        :param execution: id, Execution or ExecutionLog.
        :return: list of plots (plotly figures).
        """

        if execution is None:
            raise ValueError("execution cannot be None to retrieve plot")

        # ExecutionLog carries plot references from search_execution
        if isinstance(execution, ExecutionLog):
            if execution.plots is not None and len(execution.plots) > 0:
                plots = []
                for p in execution.plots:
                    plots.append(self.plot(plot_id=str(p.plot_id)))
                return plots
            else:
                return None
        else:
            raise TypeError("execution must be an ExecutionLog")

    def upload_model(self,
                     model_info: ModelInfo,
                     bytes_content = None):
        """
        upload a model within the model repository.
            - by default use model_info.trained_model and convert it to a pickle
            - for already torch or pickle please pass the bytes_content
            - model_info.file_format must be set properly to 'pkl' or 'pt'
        :param model_info: model info, with at least key (+twin, +property, +alias) and trained_model.
        :param bytes_content: bytes[] of your torch or pickle model.
        """
        if model_info.trained_model is None and bytes_content is None:
            raise ValueError("model_info must have a trained model (to pickle) or bytes content")
        if bytes_content is None:
            bytes_content = pickle.dumps(model_info.trained_model)
        files = {
            "trained_model": (
                "trained_model." + model_info.file_format ,
                bytes_content,
                "application/octet-stream",
            )
        }
        headers = self.__header()
        headers.pop("Content-Type", None)
        response = requests.post(self.__url() + f"models",
                                 headers=headers,
                                 data={
                                    "payload": json.dumps(model_info.to_json())
                                 },
                                 files=files)
        if response.status_code == 200:
            response_json = response.json()
            if "identifier" in response_json:
                key , twin, property_value, alias = ModelInfo.split_identifier(response_json["identifier"])
                if alias is not None:
                    model_info.alias = alias
            for file in model_info.files:
                file: wizata_dsapi.ModelFile
                if file.name != "trained_model":
                    single_file = {
                        "file": (
                            file.path,
                            file.content,
                            "application/octet-stream",
                        )
                    }
                    response = requests.post(self.__url() + f"models/{model_info.identifier(include_alias=True)}/files/{file.path}",
                                             headers=headers,
                                             files=single_file,
                                             timeout=60)
                    if response.status_code != 200 and response.status_code != 201:
                        raise self.__raise_error(response)
            return model_info
        else:
            raise self.__raise_error(response)

    def upsert_experiment(self, key: str, name: str, pipeline=None, twin=None):
        """
        Upsert an experiment.
        :param key: unique key identifying the experiment.
        :param name: display name of the experiment.
        :param pipeline: pipeline to set at creation only - cannot be updated.
        :param twin: twin uuid, hardware id or twin.
        :return: upserted experiment.
        """
        experiment = self.get(experiment_key=key)

        twin_id = None
        if twin:
            if isinstance(twin, uuid.UUID) or (isinstance(twin, str) and is_valid_uuid(twin)):
                twin_id = twin
            elif isinstance(twin, str):
                twin_id = self.get(twin_hardware_id=twin).twin_id
            elif isinstance(twin, Twin):
                twin_id = twin.twin_id

        if experiment is not None:
            found = True
            experiment.name = name
            if twin_id is not None:
                experiment.twin_id = twin_id
        else:
            experiment = Experiment(
                key=key,
                name=name
            )
            if twin_id is not None:
                experiment.twin_id = twin_id
            if pipeline:
                if isinstance(pipeline, uuid.UUID) or (isinstance(pipeline, str) and is_valid_uuid(pipeline)):
                    experiment.pipeline_id = pipeline
                elif isinstance(pipeline, str):
                    experiment.pipeline_id = self.get(pipeline_key=pipeline).pipeline_id
                elif isinstance(pipeline, Pipeline):
                    experiment.pipeline_id = pipeline.pipeline_id
            else:
                raise KeyError('please, specify a valid pipeline for an experiment creation.')
            found = False

        if not found:
            return self.create(experiment)
        else:
            return self.update(experiment)

    def upsert_template(self, key: str, name: str):
        """
        Upsert a template.
        :param key: unique key identifying the template.
        :param name: display name of the template
        :return: upserted template.
        """
        template = Template(key=key, name=name)
        return self.upsert(template)

    def upsert_pipeline(self, pipeline: Pipeline):
        """
        Upsert a pipeline (ignore ID, use the key).
        :return: upserted pipeline.
        """
        return self.upsert(pipeline)

    def upsert_datapoint(self, datapoint: DataPoint):
        """
        Upsert a datapoint (ignore ID, use the hardware_id).
        :return: upserted datapoint.
        """
        return self.upsert(datapoint)

    def upsert_twin(self, twin: Twin):
        """
        Upsert a twin (ignore ID, use the hardware_id).
        :return: upserted twin.
        """
        return self.upsert(twin)

    def add_template_property(self,
                              template,
                              property_name: str,
                              property_type: str = "datapoint",
                              is_required: bool = True) -> wizata_dsapi.TemplateProperty:

        # check type
        p_type = VarType(property_type)

        if isinstance(template, str):
            template_id = self.get(template_key=template).template_id
        elif isinstance(template, wizata_dsapi.Template):
            template_id = template.template_id
        elif isinstance(template, uuid.UUID):
            template_id = template
        else:
            raise TypeError('template should be a UUID or at least a key str referring template or a proper wizata_dsapi Template')

        template_property = wizata_dsapi.TemplateProperty(
            template_id=template_id,
            p_type=p_type,
            name=property_name,
            required=is_required
        )
        response = requests.post(self.__url() + "templateproperties/",
                                 headers=self.__header(),
                                 data=json.dumps(template_property.to_json()))
        if response.status_code == 200:
            return template_property
        else:
            raise self.__raise_error(response)

    def remove_template_property(self, template, property_name: str):

        if isinstance(template, str):
            template = self.get(template_key=template)
        elif not isinstance(template, Template):
            raise TypeError('template must be a key str referring template or a proper wizata_dsapi Template')

        if property_name is None:
            raise ValueError('please provide a valid name for the property')

        property_dict = {
            "name": property_name
        }

        response = requests.delete(self.__url() + "templates/" + str(template.template_id) + '/properties/',
                                   headers=self.__header(),
                                   data=json.dumps(property_dict))
        if response.status_code == 200:
            return
        else:
            raise self.__raise_error(response)

    def get_template_properties(self, template) -> list:
        """
        Retrieve all properties for a given template.

        :param template: template object, UUID or str key.
        :return: list of TemplateProperty.
        """
        if isinstance(template, str):
            template_id = self.get(template_key=template).template_id
        elif isinstance(template, wizata_dsapi.Template):
            template_id = template.template_id
        elif isinstance(template, uuid.UUID):
            template_id = template
        else:
            raise TypeError('template should be a UUID, a key str referring template or a proper wizata_dsapi Template')

        response = requests.get(self.__url() + "templateproperties/?templateId=" + str(template_id),
                                headers=self.__header())
        if response.status_code == 200:
            results = []
            for item in response.json():
                results.append(wizata_dsapi.TemplateProperty.from_dict(item))
            return results
        else:
            raise self.__raise_error(response)

    def get_registrations(self, template) -> list:
        """
        retrieve all registrations for
        :param template: template object, UUID or str key.
        :return: list of twin registration.
        """

        if template is None:
            raise ValueError('template must be specified.')
        elif isinstance(template, uuid.UUID):
            template = self.get(wizata_dsapi.Template(template_id=template))
        elif isinstance(template, str):
            template = self.get(template_key=template)
        elif not isinstance(template, Template):
            raise TypeError('template must be UUID, a key str referring template or a proper wizata_dsapi Template')
        if template is None:
            raise ValueError('template cannot be found on server')

        response = requests.get(self.__url() + "templates/" + str(template.template_id) + '/registrations/',
                                headers=self.__header())
        if response.status_code == 200:
            registrations_json = response.json()
            registrations = []
            for twin_registration_json in registrations_json:
                registration = TwinRegistration()
                registration.from_json(twin_registration_json)
                registrations.append(registration)
            return registrations
        else:
            raise self.__raise_error(response)

    def register_twin(self, template, twin, properties: dict, override=True):
        """
        register a twin on a specific template using a map.
        :param template: template object, UUID or str key.
        :param twin: twin object, UUID or str key.
        :param properties: dict where key = template property and value = datapoint name or const value (str, int, float, relative or epoch datetime).
        :param override: by default at True - allow overriding any existing subscription
        """

        if template is None:
            raise ValueError('template must be specified.')
        elif isinstance(template, uuid.UUID):
            template = self.get(wizata_dsapi.Template(template_id=template))
        elif isinstance(template, str):
            template = self.get(template_key=template)
        elif not isinstance(template, Template):
            raise TypeError('template must be UUID, a key str referring template or a proper wizata_dsapi Template')

        if template is None:
            raise ValueError('template cannot be found on server')

        if twin is None:
            raise ValueError('twin must be specified.')
        elif isinstance(twin, uuid.UUID):
            twin_id = twin
        elif isinstance(twin, str):
            twin_id = self.get(twin_hardware_id=twin).twin_id
        elif isinstance(twin, Twin):
            twin_id = twin.twin_id
        else:
            raise TypeError('twin must be UUID, a key str referring hardware ID or a proper wizata_dsapi Twin')

        if template.properties is None or len(template.properties) == 0:
            raise ValueError('template chosen does not contains any properties')

        if properties is None or len(properties.keys()) == 0:
            raise ValueError('your map dictionary must contains properties matching template')

        twin_properties = {
            "properties": []
        }
        for key in properties.keys():
            property_type = None
            for template_property in template.properties:
                template_property: TemplateProperty
                if key == template_property.name:
                    property_type = template_property.p_type
            if property_type is None:
                raise ValueError(f'cannot find property {key} in template or cannot determine its type')
            twin_properties["properties"].append({
                'name': key,
                property_type.value: properties[key]
            })

        # Check if already exist
        exists = False
        response_exists = requests.get(self.__url() + "templates/" + str(template.template_id)
                                       + '/registrations/' + str(twin_id) + '/',
                                       headers=self.__header())
        if response_exists.status_code == 200:
            exists = True

        if exists and not override:
            raise ValueError("registration already exists and override not allowed.")

        if exists and override:
            response = requests.put(self.__url() + "templates/" + str(template.template_id)
                                    + '/registrations/' + str(twin_id) + '/',
                                    headers=self.__header(),
                                    data=json.dumps(twin_properties))
        else:
            response = requests.post(self.__url() + "templates/" + str(template.template_id)
                                     + '/registrations/' + str(twin_id) + '/',
                                     headers=self.__header(),
                                     data=json.dumps(twin_properties))

        if response.status_code == 200:
            return
        else:
            raise self.__raise_error(response)

    def unregister_twin(self, template, twin):
        """
        un-register a twin from a specific template.
        :param template: template object, UUID or str key.
        :param twin: twin object, UUID or str key.
        """

        if template is None:
            raise ValueError('template must be specified.')
        elif isinstance(template, uuid.UUID):
            template = self.get(wizata_dsapi.Template(template_id=template))
        elif isinstance(template, str):
            template = self.get(template_key=template)
        elif not isinstance(template, Template):
            raise TypeError('template must be UUID, a key str referring template or a proper wizata_dsapi Template')

        if twin is None:
            raise ValueError('twin must be specified.')
        elif isinstance(twin, uuid.UUID):
            twin_id = twin
        elif isinstance(twin, str):
            twin_id = self.get(twin_hardware_id=twin).twin_id
        elif isinstance(twin, Twin):
            twin_id = twin.twin_id
        else:
            raise TypeError('twin must be UUID, a key str referring hardware ID or a proper wizata_dsapi Twin')

        response = requests.delete(self.__url() + "templates/" + str(template.template_id)
                                   + '/registrations/' + str(twin_id) + '/',
                                   headers=self.__header())
        if response.status_code == 200:
            return
        else:
            raise self.__raise_error(response)

    def create_component(self, component: SolutionComponent):
        """
        create a component based on its ID.
        """
        if component is None:
            raise ValueError("component cannot be null")

        response = requests.post(self.__url() + "components/",
                                 headers=self.__header(),
                                 data=json.dumps(component.to_json()))

        if response.status_code == 200:
            return
        else:
            raise self.__raise_error(response)

    def update_component(self, component: SolutionComponent):
        """
        update a component based on its ID.
        """
        if component is None:
            raise ValueError("component cannot be null")

        response = requests.put(self.__url() + "components/" + str(component.solution_component_id) + "/",
                                 headers=self.__header(),
                                 data=json.dumps(component.to_json()))

        if response.status_code == 200:
            return
        else:
            raise self.__raise_error(response)

    def get_datapoint_mappings(self, registration):
        """
        get datapoint mapping from a registration.
        """
        if isinstance(registration, uuid.UUID):
            registration = str(registration)
        elif isinstance(registration, TwinRegistration):
            registration = str(registration.twin_registration_id)
        response = requests.request("GET",
                                    self.__url() + "registrations/" + registration + "/mapping/",
                                    headers=self.__header()
                                    )
        if response.status_code == 200:
            return response.json()
        else:
            raise self.__raise_error(response)

    def get_components(self,
                       label_id: uuid.UUID = None,
                       twin_id: uuid.UUID = None,
                       template_id: uuid.UUID = None,
                       owner_id: uuid.UUID = None,
                       organization_only: bool = False,
                       name: str = None):
        """
        get components
        :param label_id: filter on a specific label
        :param template_id: filter on a specific template
        :param twin_id: filter on a specific twin
        :param owner_id: filter on a specific owner_id
        :param organization_only: work only with organization components (by default - False)
        :param name: filter on a specific name (contains)
        """
        params = {}
        if label_id is not None:
            params['labelId'] = str(label_id)
        if twin_id is not None:
            params['twinId'] = str(twin_id)
        if template_id is not None:
            params['templateId'] = str(template_id)
        if owner_id is not None:
            params['ownerId'] = str(owner_id)
        if organization_only:
            params['organizatianOnly'] = True
        if name is not None:
            params['name'] = str(name)
        response = requests.request("GET",
                                    self.__url() + "components/",
                                    params=params,
                                    headers=self.__header()
                                    )
        if response.status_code == 200:
            components = []
            for json_model in response.json():
                component = SolutionComponent()
                component.from_json(json_model)
                components.append(component)
            return components
        else:
            raise self.__raise_error(response)

    def delete_component(self, component_id: uuid.UUID):
        response = requests.delete(self.__url() + "components/" + str(component_id) + '/',
                                   headers=self.__header())
        if response.status_code == 200:
            return
        else:
            raise self.__raise_error(response)

    def search_datapoints(self,
                          page: int = 1,
                          size: int = 20,
                          sort: str = "id",
                          direction: str = "asc",
                          hardware_id: str = None,
                          categories: list = None,
                          business_types: list = None,
                          twin=None,
                          recursive: bool = False) -> PagedQueryResult:
        """
        get datapoints with a paged query.
        :param page: numero of the page - default 1.
        :param size: quantity per page - default 20 max 100.
        :param sort: column to sort results - default id.
        :param direction: sorting direction by default asc, accept also desc.
        :param hardware_id: filter on a specific hardware ID name or partial name.
        :param business_types: list of BusinessType or str.
        :param categories: list of UUID or Category.
        :param twin: uuid or Twin element to search datapoints.
        :param recursive: set to True in combination of a twin to look inside all sub-twins recursively.
        :return: PagedQueryResults, check total for number of potential results and results for the list of entity.
        """
        query = PagedQueryResult(
            page=page,
            size=size,
            sort=sort,
            direction=direction
        )

        if sort not in ["id", "hardwareId", "createdDate", "createdById", "updatedDate", "updatedById",
                        "unitId", "businessType", "categoryId", "description"]:
            raise ValueError("invalid sort column")

        parameter_str = "?page=" + str(page) + "&size=" + str(size) + "&sort=" + str(sort) + "&direction=" + str(direction)
        if hardware_id is not None:
            parameter_str += "&hardwareId=" + sanitize_url_parameter(hardware_id)
        if business_types is not None and len(business_types) > 0:
            list_str = []
            for businessType in business_types:
                if isinstance(businessType, str):
                    list_str.append(businessType)
                elif isinstance(businessType, BusinessType):
                    list_str.append(businessType.value)
                else:
                    raise ValueError('business type must be a BusinessType or a str')
            parameter_str += "&businessTypes=" + sanitize_url_parameter(",".join(list_str))
        if categories is not None and len(categories) > 0:
            list_str = []
            for categoryId in categories:
                if isinstance(categoryId, uuid.UUID):
                    list_str.append(str(categoryId))
                if isinstance(categoryId, str):
                    list_str.append(categoryId)
                elif isinstance(categoryId, Category):
                    list_str.append(str(categoryId.category_id))
                else:
                    raise ValueError('category must be a Category, UUID or a str UUID')
            parameter_str += "&categoryIds=" + sanitize_url_parameter(",".join(list_str))

        if recursive is not None and recursive:
            parameter_str += "&recursive=" + sanitize_url_parameter("true")

        if twin is not None:
            if isinstance(twin, uuid.UUID):
                twin = str(twin)
            if isinstance(twin, str):
                twin = twin
            elif isinstance(twin, Twin):
                twin = str(twin.twin_id)
            else:
                raise ValueError('category must be a Twin, UUID or a str UUID')
            parameter_str += "&twinId=" + sanitize_url_parameter(twin)

        response = requests.request("GET",
                                    self.__url() + "datapoints/filter" + parameter_str,
                                    headers=self.__header()
                                    )
        if response.status_code == 200:
            response_obj = response.json()
            query.total = int(response_obj["total"])
            query.results = []
            for obj in response_obj["results"]:
                datapoint = DataPoint()
                datapoint.from_json(obj)
                query.results.append(datapoint)
            return query
        else:
            raise self.__raise_error(response)

    def search_twins(self,
                     page: int = 1,
                     size: int = 20,
                     sort: str = "id",
                     direction: str = "asc",
                     hardware_id: str = None,
                     name: str = None,
                     parents: list = None) -> PagedQueryResult:
        """
        get twins with a paged query.
        :param page: numero of the page - default 1.
        :param size: quantity per page - default 20 max 100.
        :param sort: column to sort results - default id.
        :param direction: sorting direction by default asc, accept also desc.
        :param hardware_id: filter on a specific hardware ID name or partial name.
        :param name: name or part of twin name.
        :param parents: list of all possible parents (Twin, UUID, or str UUID).
        :return: PagedQueryResults, check total for number of potential results and results for the list of entity.
        """
        query = PagedQueryResult(
            page=page,
            size=size,
            sort=sort,
            direction=direction
        )

        if sort not in ["id", "hardwareId", "createdDate", "createdById", "updatedDate", "updatedById",
                        "name", "description"]:
            raise ValueError("invalid sort column")

        parameter_str = "?page=" + str(page) + "&size=" + str(size) + "&sort=" + str(sort) + "&direction=" + str(direction)
        if hardware_id is not None:
            parameter_str += "&hardwareId=" + sanitize_url_parameter(hardware_id)
        if parents is not None and len(parents) > 0:
            list_str = []
            for parent in parents:
                if isinstance(parent, uuid.UUID):
                    list_str.append(str(parent))
                elif isinstance(parent, str):
                    list_str.append(parent)
                elif isinstance(parent, Twin):
                    list_str.append(str(parent.twin_id))
                else:
                    raise ValueError('parent must be a Twin or a str')
            parameter_str += "&parentIds=" + sanitize_url_parameter(",".join(list_str))

        if name is not None:
            parameter_str += "&name=" + sanitize_url_parameter(name)

        response = requests.request("GET",
                                    self.__url() + "twins/filter" + parameter_str,
                                    headers=self.__header()
                                    )
        if response.status_code == 200:
            response_obj = response.json()
            query.total = int(response_obj["total"])
            query.results = []
            for obj in response_obj["results"]:
                twin = Twin()
                twin.from_json(obj)
                query.results.append(twin)
            return query
        else:
            raise self.__raise_error(response)

    def search_insights(self,
                        page: int = 1,
                        size: int = 20,
                        sort: str = "id",
                        direction: str = "asc",
                        datapoint_id: str = None,
                        twin_id: str = None,
                        component_id: list = None):
        """
        get insights with a paged query.
        :param page: numero of the page - default 1.
        :param size: quantity per page - default 20 max 100.
        :param sort: column to sort results - default id.
        :param direction: sorting direction by default asc, accept also desc.
        :param datapoint_id: id of the datapoint on which filtering the insights.
        :param twin_id: id of the twin on which filtering the insights.
        :param component_id: id of the component on which filtering the insights.
        :return: PagedQueryResults, check total for number of potential results and results for the list of entity.
        """
        query = PagedQueryResult(
            page=page,
            size=size,
            sort=sort,
            direction=direction
        )

        if sort not in ["id", "sensorId", "twinId", "componentId",
                        "name", "displayPrecision"]:
            raise ValueError("invalid sort column")

        parameter_str = "?page=" + str(page) + "&size=" + str(size) + "&sort=" + str(sort) + "&direction=" + str(direction)

        if datapoint_id is not None:
            parameter_str += "&datapointId=" + sanitize_url_parameter(str(datapoint_id))

        if twin_id is not None:
            parameter_str += "&twinId=" + sanitize_url_parameter(str(twin_id))

        if component_id is not None:
            parameter_str += "&componentId=" + sanitize_url_parameter(str(component_id))

        response = requests.request("GET",
                                    self.__url() + "insights/filter" + parameter_str,
                                    headers=self.__header()
                                    )
        if response.status_code == 200:
            response_obj = response.json()
            query.total = int(response_obj["total"])
            query.results = []
            for obj in response_obj["results"]:
                insight = Insight()
                insight.from_json(obj)
                query.results.append(insight)
            return query
        else:
            raise self.__raise_error(response)

    def search_executions(self,
                          page: int = 1,
                          size: int = 20,
                          sort: str = "id",
                          direction: str = "asc",
                          pipeline_id: uuid.UUID = None,
                          twin_id:  uuid.UUID = None,
                          template_id: uuid.UUID = None,
                          status: ExecutionStatus = None) -> PagedQueryResult:
        """
        get executions with a paged query.
        :param page: numero of the page - default 1.
        :param size: quantity per page - default 20 max 100.
        :param sort: column to sort results - default id.
        :param direction: sorting direction by default asc, accept also desc.
        :param pipeline_id: filter on a specific pipeline.
        :param twin_id: filter on a specific twin.
        :param template_id: filter on a specific template.
        :param status: filter on a specific status.
        :return: PagedQueryResults, check total for number of potential results and results for the list of entity.
        """
        query = PagedQueryResult(
            page=page,
            size=size,
            sort=sort,
            direction=direction
        )

        if sort not in ["id", "pipelineId", "twinId", "templateId",
                        "createdDate" , "createdById",
                        "updatedDate", "updatedById",
                        "status"]:
            raise ValueError("invalid sort column")

        parameter_str = "?page=" + str(page) + "&size=" + str(size) + "&sort=" + str(sort) + "&direction=" + str(direction)

        if pipeline_id is not None:
            parameter_str += "&pipelineId=" + sanitize_url_parameter(str(pipeline_id))

        if twin_id is not None:
            parameter_str += "&twinId=" + sanitize_url_parameter(str(twin_id))

        if template_id is not None:
            parameter_str += "&templateId=" + sanitize_url_parameter(str(template_id))

        if status is not None:
            if not isinstance(status, ExecutionStatus):
                raise ValueError(f'status is not a valid ExecutionStatus')
            parameter_str += "&status=" + sanitize_url_parameter(status.value)

        response = requests.request("GET",
                                    self.__url() + "execute/filter" + parameter_str,
                                    headers=self.__header()
                                    )
        if response.status_code == 200:
            response_obj = response.json()
            query.total = int(response_obj["total"])
            query.results = []
            for obj in response_obj["results"]:
                execution = Execution()
                execution.from_json(obj)
                query.results.append(execution)
            return query
        else:
            raise self.__raise_error(response)

    def search_execution(self, search_query=None) -> PagedQueryResult:
        """
        search executions using filters, sorting and pagination via POST.
        :param search_query: dict or SearchQuery with filters, sort, page, size, groups, groupSummaries, serviceName.
        :return: PagedQueryResult with ExecutionLog results and optional group statistics.
        """
        if search_query is None:
            search_query = {}

        if isinstance(search_query, SearchQuery):
            search_query = search_query.to_dict()

        response = requests.post(
            self.__url() + "execute/search",
            headers=self.__header(),
            data=json.dumps(search_query)
        )
        if response.status_code == 200:
            return PagedQueryResult.from_dict(response.json(), dto_class=ExecutionLog)
        else:
            raise self.__raise_error(response)

    def search_models(self) -> ModelList:
        """
        get all information related to models stored on Wizata.
        :return: ModelList structure model list
        """
        response = requests.request("GET",
                                    self.__url() + "models",
                                    headers=self.__header()
                                    )
        if response.status_code == 200:
            response_json = response.json()
            model_list = ModelList()
            for model_json in response_json:
                model_info = ModelInfo(model_json["key"])
                model_info.from_json(model_json)
                model_info._api = self
                model_list.append(model_info)
            return model_list
        else:
            raise self.__raise_error(response)

    def abort(self, executions: list) -> str:
        """
        send an abort request for executions and return a result message
        :param executions: must be a list containing uuid or Execution.
        """
        if executions is None or len(executions) == 0:
            raise ValueError(f'please provide at list one execution to abort')

        payload = {
            'ids': []
        }
        for execution in executions:
            if isinstance(execution, uuid.UUID):
                payload['ids'].append(str(execution))
            if isinstance(execution, str):
                payload['ids'].append(str(uuid.UUID(execution)))
            if isinstance(execution, Execution):
                payload['ids'].append(str(execution.execution_id))
        response = requests.post(
            self.__url() + "execute/abort",
            headers=self.__header(),
            data=json.dumps(payload)
        )
        return response.text

    def build_image(self, key: str) -> str:
        """
        build an image of a pipeline, store it on the image repository and return its pipeline image id.
        :param key: pipeline key.
        :return: image id.
        """
        response = self._process_request(
            method="POST",
            route=PipelineImage.route() + "/",
            dto_class=PipelineImage,
            data=json.dumps({
                'key': key
            })
        )
        return response.json()["id"]

    def download_image(self, pipeline_image_id: str) -> PipelineImage:
        """
        download a pipeline image from the repository
        :param pipeline_image_id: id of image
        :return: packaged (unzipped) as a PipelineImage
        """

        response_bytes = self._process_request(
            method="GET",
            route=PipelineImage.route() + "/" + str(pipeline_image_id) + "/download/",
            result_type="bytes"
        )
        image = PipelineImage.loads(pipeline_image_id=pipeline_image_id, g_bytes=response_bytes)
        return image

    def download_model(self, identifier: str) -> ModelInfo:
        """
        download a model image directly from the repository, auto-select alias if necessary.
        :param identifier: exact identifier including alias of the model or no alias to auto-select active version.
        :return: ModelInfo with trained_model loaded, extra file must be downloaded separately.
        """
        response = requests.get(self.__url() + f"models/{identifier}/",
                                headers=self.__header())
        if response.status_code == 200:
            model_info = ModelInfo()
            model_info.from_json(response.json())
            self.load_model(model_info)
            return model_info
        else:
            raise self.__raise_error(response)


    def delete_model(self, identifier: str):
        """
        delete a model version from the repository.
        :param identifier: full identifier including alias (e.g. 'my_model@alias' or 'my_model.twin@alias').
        """
        response = requests.delete(self.__url() + f"models/{identifier}/",
                                   headers=self.__header())
        if response.status_code != 200:
            raise self.__raise_error(response)

    def download_file(self,
                      model: ModelInfo = None,
                      file: ModelFile = None,
                      identifier: str = None,
                      path: str = None):
        """
        download a model extra file from the repository.
        :param model: model (ModelInfo), alternatively can use identifier
        :param file: file (ModelFile), alternatively can use path
        :param identifier: identifier including alias of the model.
        :param path: path name of the file including the extension.
        :return: bytes content.
        """

        if model is None and identifier is None:
            raise TypeError('model or identifier must be provided')

        if file is None and path is None:
            raise TypeError('file or path must be provided')

        if identifier is None:
            identifier = model.identifier(include_alias=True)

        if path is None:
            path = file.path

        response = requests.get(self.__url() + f"models/{identifier}/files/{path}/",
                                headers=self.__header())
        if response.status_code == 200:
            return response.content
        else:
            self.__raise_error(response)


    def upload_file(self,
                    content,
                    model: ModelInfo = None,
                    file: ModelFile = None,
                    identifier: str = None,
                    path: str = None
                    ):
        """
        upload a model extra file to the repository.
        please use upload_model for a trained_model
        :param content: the bytes content of the file.
        :param model: model (ModelInfo), alternatively can use identifier
        :param file: file (ModelFile), alternatively can use path
        :param identifier: identifier including alias of the model.
        :param path: path name of the file including the extension.
        :return: bytes content.
        """
        if path == "trained_model.pkl" or path == "trained_model.pt" or path == "_metadata.json":
            raise RuntimeError("upload_file cannot be used to upload reserved metadata or the primary trained model")

        if model is None and identifier is None:
            raise TypeError('model or identifier must be provided')

        if file is None and path is None:
            raise TypeError('file or path must be provided')

        if identifier is None:
            identifier = model.identifier(include_alias=True)

        if path is None:
            path = file.path

        headers = self.__header()
        headers.pop("Content-Type", None)
        single_file = {
            "file": (
                path,
                content,
                "application/octet-stream",
            )
        }
        response = requests.post(self.__url() + f"models/{identifier}/files/{path}",
                                 headers=headers,
                                 files=single_file,
                                 timeout=60)
        if response.status_code != 200 and response.status_code != 201:
            return
        else:
            self.__raise_error(response)

    def load_model(self, model):
        """
        load a model pickle or torch from the repository ready to be used.
        :param model: ModelInfo to load
        :return: ModelInfo with the trained model.
        """
        if not isinstance(model, ModelInfo):
            raise TypeError('model must be an instance of ModelInfo')

        identifier = model.identifier(include_alias=True)
        extension = model.file_format
        response = requests.get(self.__url() + f"models/{identifier}/files/trained_model.{extension}/",
                                headers=self.__header())
        if response.status_code == 200:
            if extension == 'pkl':
                model.trained_model = joblib.load(io.BytesIO(response.content))
                return model
            elif extension == 'pt':
                import torch
                model.trained_model = torch.jit.load(io.BytesIO(response.content))
                return model
            else:
                raise ValueError(f'unsupported file format {extension}')
        else:
            self.__raise_error(response)


def api() -> WizataDSAPIClient:
    """
    Create a WizataDSAPIClient from environment variables.
    :return: client
    """
    protocol = 'https'

    streamlit_domain = get_streamlit_domain()
    if streamlit_domain is not None:
        return WizataDSAPIClient(
            domain=streamlit_domain,
            protocol=protocol
        )

    if os.environ.get('WIZATA_CLIENT_ID') is None:
        raise ValueError('please configure OS variables to authenticate with api() method')

    if os.environ.get('WIZATA_PROTOCOL') is not None:
        protocol = os.environ.get('WIZATA_PROTOCOL')

    if os.environ.get('WIZATA_CLIENT_SECRET') is not None:
        scope = os.environ.get('WIZATA_SCOPE')
        if scope is None:
            scope = os.environ.get('WIZATA_CLIENT_ID') + '/.default'
        return WizataDSAPIClient(
            tenant_id=os.environ.get('WIZATA_TENANT_ID'),
            client_id=os.environ.get('WIZATA_CLIENT_ID'),
            client_secret=os.environ.get('WIZATA_CLIENT_SECRET'),
            scope=scope,
            domain=os.environ.get('WIZATA_DOMAIN'),
            protocol=protocol
        )
    else:
        return WizataDSAPIClient(
            tenant_id=os.environ.get('WIZATA_TENANT_ID'),
            client_id=os.environ.get('WIZATA_CLIENT_ID'),
            scope=os.environ.get('WIZATA_SCOPE'),
            username=os.environ.get('WIZATA_USERNAME'),
            password=os.environ.get('WIZATA_PASSWORD'),
            domain=os.environ.get('WIZATA_DOMAIN'),
            protocol=protocol
        )


def is_valid_uuid(val):
    try:
        uuid.UUID(str(val))
        return True
    except ValueError:
        return False



import pandas
import numpy
import io
from datetime import datetime, timedelta, timezone
import re
import random
from .words import animals, colors


def validate(df: pandas.DataFrame) -> pandas.DataFrame:
    """
    validate dataframe format to match supported Wizata format.
    :param df: dataframe to validate.
    :return: validated and formatted df - raise error if not valid.
    """
    if isinstance(df.index, pandas.MultiIndex):
        for name in df.index.names:
            if name not in ['Timestamp', 'eventId', 'eventStatus', None]:
                raise ValueError(f'illegal component of dataframe multi-index {name} not Timestamp or eventId')
    elif df.index.name == "eventId":
        pass
    elif "field" in df.columns:
        return df
    else:
        if df.empty:
            return df

        if not isinstance(df.index, pandas.DatetimeIndex):
            raise TypeError(f'df.index is not a DatatimeIndex {df.index}')

        if df.index.name != "Timestamp":
            df.index.name = "Timestamp"

    if len(df.axes) < 2 or df.axes[1].name != "sensorId":
        df.rename_axis("sensorId", axis="columns", inplace=True)

    # df = df.astype(float)

    return df


def df_to_csv(df: pandas.DataFrame) -> bytes:
    """
    Convert the DataFrame to a strongly formatted CSV.
    :param df: pandas DataFrame compatible with Wizata standards.
    :return: bytes containing the full CSV file.
    """
    b_buf = io.BytesIO()

    df.to_csv(b_buf,
              date_format="%Y-%m-%d-%H-%M-%S-%f",
              sep=",",
              decimal=".",
              encoding="utf-8")

    b_buf.seek(0)
    return b_buf.read()


def df_from_csv(b_data: bytes) -> pandas.DataFrame:
    """
    Convert the bytes to a pandas.DataFrame.
    :param b_data: bytes representing a CSV file.
    :return: pandas DataFrame formatted.
    """
    b_buf = io.BytesIO(b_data)

    df = pandas.read_csv(b_buf,
                         sep=",",
                         decimal=".",
                         encoding="utf-8")

    # detect timestamp column
    if "timestamp" in df.columns:
        df = df.rename(columns={'timestamp': 'Timestamp'})
    if "Timestamp" not in df.columns:
        raise ValueError('Cannot read dataframe as no Timestamp columns exists.')

    # detect timestamp type
    if df['Timestamp'].dtypes == 'int64':
        df['Timestamp'] = pandas.to_datetime(df['Timestamp'], unit="ms")
    elif df['Timestamp'].dtypes == 'object':
        df['Timestamp'] = df['Timestamp'].apply(lambda _: datetime.strptime(_, "%Y-%m-%d-%H-%M-%S-%f"))

    df = df.set_index('Timestamp')
    df.rename_axis("sensorId", axis="columns", inplace=True)

    return df


def df_from_json(json):
    """
    Convert a dictionary dataframe using JSON convention into a panda Dataframe.

    Dataframe must contain a timestamp column and be compatible to float data types.

    :param json: JSON formatted dataframe.
    :return: panda Dataframe
    """
    df = pandas.DataFrame.from_dict(json, orient='columns')
    df = df.set_index('Timestamp')

    if df.index.dtype == 'int64':
        df.index = [datetime.fromtimestamp(i) for i in (df.index / 1000).astype(int)]
        df.index.name = 'Timestamp'
    if not isinstance(df.index, pandas.DatetimeIndex):
        raise TypeError("Unexpected type {0}".format(df.index))

    df.rename_axis("sensorId", axis="columns", inplace=True)
    return df


def df_to_json(df: pandas.DataFrame):
    """
    Convert a panda Dataframe to a JSON compatible dictionary.

    Dataframe must be compatible to Wizata format using Timestamp index and float data types.

    :param df: panda Dataframe to convert.
    :return: dictionary representing JSON compatible dataframe.
    """
    df_json = {
        "Timestamp": list(df.index)
    }
    for col in list(df.columns):
        if col != 'Timestamp':
            df_json[col] = list(df[col].values.astype(float))
        else:
            df_json[col] = list(df[col].values)
    return df_json


def df_from_dict(df_dict: dict) -> pandas.DataFrame:
    """
    convert a dict into a valid Wizata dataframe.
    :param df_dict: dataframe dict.
    :return: pandas DataFrame
    """

    for key in df_dict:
        if isinstance(df_dict[key], float):
            df_dict[key] = [df_dict[key]]
        elif isinstance(df_dict[key], int):
            df_dict[key] = [df_dict[key]]
        elif isinstance(df_dict[key], list):
            if len(df_dict[key]) != 1:
                if "Timestamp" not in df_dict:
                    raise ValueError('if using a multi-line dataframe please provide a Timestamp column')

    if "Timestamp" not in df_dict:
        df_dict["Timestamp"] = [datetime.now(timezone.utc)]
    else:
        timestamp_col = []
        for timestamp_ms in df_dict["Timestamp"]:
            timestamp_s = timestamp_ms / 1000.0
            timestamp_col.append(datetime.fromtimestamp(timestamp_s))
        df_dict["Timestamp"] = timestamp_col

    df = pandas.DataFrame.from_dict(df_dict)
    df = df.set_index("Timestamp")
    df = df.rename_axis("sensorId", axis="columns")
    df = df.astype(float)
    return df


def df_to_dict(df: pandas.DataFrame, format_str: str = "default"):
    """
    convert a DataFrame to a dict
    :param df: dataframe to format
    :param format_str: format to use - default or grafana
    :return:
    """
    if format_str == "default":
        df = df.reset_index()
        df['Timestamp'] = df['Timestamp'].apply(lambda x: int(x.timestamp() * 1000))
        df = df.replace({numpy.nan: None})
        df_dict = df.to_dict()
        return df_dict
    elif format_str == "grafana":
        df = df.reset_index()
        df = df.replace({numpy.nan: None})
        df_dict = df.to_dict(orient="records")
        for d in df_dict:
            d["Timestamp"] = int(d["Timestamp"].timestamp() * 1000)
        return df_dict
    else:
        raise ValueError(f'please set a valid format "default","grafana"')


def verify_relative_datetime(formatted_string: str) -> bool:
    """
    verify format of a relative datetime str
    """
    if not isinstance(formatted_string, str):
        raise ValueError(f'please provide a string as relative datetime')

    if formatted_string == '':
        return True

    pattern = r'^now([+-]\d+([yMwdHhms]{1,2}))?$'
    match = re.match(pattern, formatted_string)
    if not match:
        raise ValueError(f"invalid time delay format {formatted_string}")

    return True


def generate_epoch(formatted_string: str, now=None):
    """
    generate an epoch based on a formatted string (e.g. now+6h) - see documentation.
    * now = datetime.now(timezone.utc) can be override
    * units = 'y' = 365d, 'M'=30d , 'w'=7d , 'd'=24h , 'h'=60m, 'm'=60s , 's'=1000'ms'
    * operators = '+' or '-'
    :param formatted_string: formatted epoch representation using relative time.
    :param now: override now datetime.
    :return: epoch in ms.
    """
    if now is None:
        now = datetime.now(timezone.utc)

    pattern = r'^now([+-]\d+([yMwdHhms]{1,2}))?$'
    match = re.match(pattern, formatted_string)
    if not match:
        raise ValueError(f"invalid time delay format {formatted_string}")

    operator = match.group(1)[0] if match.group(1) else None
    value_unit_str = match.group(1)[1:] if match.group(1) else None

    if value_unit_str:
        value_str = re.search(r'\d+', value_unit_str).group()
        unit_str = re.search(r'[yMwdHhms]{1,2}', value_unit_str).group()

        value = int(value_str)

        if unit_str == 'y':
            delta = timedelta(days=365) * value
        elif unit_str == 'M':
            delta = timedelta(days=30) * value
        elif unit_str == 'w':
            delta = timedelta(days=7) * value
        elif unit_str == 'd':
            delta = timedelta(days=value)
        elif unit_str == 'H' or unit_str == 'h':
            delta = timedelta(hours=value)
        elif unit_str == 'm':
            delta = timedelta(minutes=value)
        elif unit_str == 's':
            delta = timedelta(seconds=value)
        elif unit_str == 'ms':
            delta = timedelta(milliseconds=value)
        else:
            raise ValueError("invalid time unit")

        if operator == '+':
            timestamp = now + delta
        elif operator == '-':
            timestamp = now - delta
        else:
            raise ValueError("invalid time delay format")
    else:
        timestamp = now

    epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)  # aware
    timestamp = timestamp.astimezone(timezone.utc)
    timestamp_ms = int((timestamp - epoch).total_seconds() * 1000)
    return timestamp_ms


def generate_unique_key() -> str:
    """
    generate a unique key for experiment, pipeline, model, template, ...
        - 7 char reserved for 'date_'
        - 11 char for 'color_'
        - 11 char for 'animal_'
        - 3 car for number XXX
    """
    return f"{datetime.now(timezone.utc).strftime('%y%m%d')}_{random.choice(colors).lower()}_" \
           f"{random.choice(animals).lower()}_{random.randint(1, 999)}"



from setuptools import setup
from distutils.util import convert_path

main_ns = {}
ver_path = convert_path('wizata_dsapi/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

core_dependencies = [
    'pandas==1.5.3; python_version<"3.11"',
    'pandas==2.2.3; python_version>="3.11" and python_version<"3.12"',
    'pandas>=2.3.3; python_version>="3.12"',
    'numpy==1.26.4; python_version<"3.12"',
    'numpy>=2.3.3; python_version>="3.12"',
    'dill==0.3.6; python_version<"3.12"',
    'dill==0.4.0; python_version>="3.12"',
    'protobuf==3.20.3; python_version<"3.12"',
    #'protobuf==5.29.5; python_version>="3.11" and python_version<"3.12"',
    'protobuf>=5.29.5; python_version>="3.12"',
    'msal==1.30.0; python_version<"3.12"',
    'msal>=1.34.0; python_version>="3.12"',
    'joblib==1.5.2; python_version<"3.12"',
    'joblib>=1.2.0; python_version>="3.12"',
    'requests==2.28.2; python_version<"3.11"',
    'requests>=2.32.5; python_version>="3.11"',
    'setuptools==67.6.0; python_version<"3.11"',
    'setuptools>=80.9.0; python_version>="3.11"',
    'plotly==5.13.1; python_version<"3.12"',
    'plotly>=6.3.1; python_version>="3.12"',
    'pytz>=2026.1'
]

ml_dependencies = [
    'matplotlib==3.7.1; python_version<"3.11"',
    'matplotlib>=3.10.6; python_version>="3.11"',
    #'tensorflow>=2.15.0; (sys_platform != "darwin" or platform_machine != "arm64") and python_version<"3.12"',
    #'tensorflow-macos>=2.15.0; sys_platform == "darwin" and platform_machine == "arm64" and python_version<"3.12"',
    #'tensorflow>=2.20.0; python_version>="3.12"',
    #'keras==2.15.0; python_version<"3.12"',
    #'keras>=3.11.3; python_version>="3.12"',
    #'tensorflow_probability==0.15.0; python_version<"3.12"',
    #'tensorflow_probability>=0.25.0; python_version>="3.12"',
    'scikit-learn==1.2.2; python_version<"3.12"',
    'scikit-learn>=1.7.2; python_version>="3.12"',
    'adtk==0.6.2',
    'scipy==1.10.1; python_version<"3.12"',
    'scipy>=1.16.2; python_version>="3.12"',
    'xgboost==1.7.4; python_version<"3.12"',
    'xgboost>=3.0.5; python_version>="3.12"',
    'u8darts==0.25.0; python_version<"3.12"',
    'u8darts>=0.38.0; python_version>="3.12"',
    'optuna==3.3.0; python_version<"3.12"',
    'optuna>=4.5.0; python_version>="3.12"',
    'explainerdashboard==0.4.2.1; python_version<"3.12"',
    'ipywidgets==8.0.4; python_version<"3.12"',
    'ipywidgets>=8.1.7; python_version>="3.12"',
    'kaleido==0.2.1; python_version<"3.12"',
    'kaleido>=1.1.0; python_version>="3.12"',
    'pytest==7.2.2; python_version<"3.12"',
    'pytest>=8.4.2; python_version>="3.12"',
    'pytest-cov==4.0.0; python_version<"3.12"',
    'pytest-cov>=7.0.0; python_version>="3.12"',
    'shapely==2.0.1; python_version<"3.12"',
    'shapely>=2.1.2; python_version>="3.12"',
    'pyodbc==4.0.35; python_version<"3.12"',
    'pyodbc>=5.2.0; python_version>="3.12"',
    'torch==2.7.1; python_version>="3.11" and python_version<"3.12"',
    'torch>=2.8.0; python_version>="3.12"'
]
all_dependencies = core_dependencies + ml_dependencies

setup(
    name='wizata_dsapi',
    version=main_ns['__version__'],
    description='Wizata Data Science Toolkit',
    author='Wizata S.A.',
    author_email='info@wizata.com',
    packages=['wizata_dsapi',
              'wizata_dsapi.plots',
              'wizata_dsapi.scripts',
              'wizata_dsapi.models'],
    install_requires=core_dependencies,
    extras_require={
        'all': all_dependencies
    },
)



from .api_dto import Dto
from .plot import Plot
from .experiment import Experiment
from .pipeline import Pipeline
from .twin import Twin
from .template import Template
from .twinregistration import TwinRegistration
from .execution import ExecutionStatus
from typing import Optional, Union
import uuid
import sys
import json


class ExecutionLog(Dto):
    """
    execution log contains all information about a current execution.
    - execution log replaces execution in version 11.4+
    - as based on loki storage, some data are values and others are indexes/labels
    - other properties are used only during execution time and are non-persistent

    # labels
    :ivar str experiment: short string key identifier of the experiment
    :ivar str pipeline: short string key identifier of the pipeline
    :ivar str template: short string key identifier of the template
    :ivar str twin: short string key identifier of the twin
    :ivar str edge_device_id: short string id of the IoT Edge Device
    :ivar wizata_dsapi.ExecutionStatus status: execution status
    :ivar str level: logging level for the execution (info, warning, error)

    # values
    :ivar uuid.UUID id: technical id of execution
    :ivar str pipeline_image_id: pipeline image identifier
    :ivar int timestamp: timestamp on which execution was created (in milliseconds)
    :ivar int started_date: timestamp on which execution is started
    :ivar int queued_date: timestamp on which execution is queued, if none use createdDate
    :ivar int waiting_time: duration of the waiting time in the queue
    :ivar int execution_time: duration of the execution from started
    :ivar uuid.UUID trigger_id: technical id of trigger
    :ivar str version: python version (major.minor) to use as target on runners, by default use current auto-detected version.


    # non-persistent values
    :ivar dict properties: properties containing key/value pairs used to parametrize the pipeline execution
    :ivar list messages: list of error and/or warning messages
    """

    @classmethod
    def from_dict(cls, data: dict) -> "ExecutionLog":

        # un-pack properties if they are jsonify
        properties = data.get("properties")
        if isinstance(properties, str):
            properties = json.loads(properties)

        plots = data.get("plots")
        _plots = []
        if isinstance(plots, list):
            for plot in plots:
                _plots.append(Plot.from_dict(plot))

        return cls(
            execution_id=uuid.UUID(data["id"]) if "id" in data and not isinstance(data["id"], int) else None,
            experiment=data.get("experiment"),
            pipeline=data.get("pipeline") if "pipeline" in data and data["pipeline"] is not None else uuid.UUID(data.get("pipelineId")) if "pipelineId" in data and data["pipelineId"] is not None else None,
            template=data.get("template") if "template" in data and data["template"] is not None else uuid.UUID(data.get("templateId")) if "templateId" in data and data["templateId"] is not None else None,
            twin=data.get("twin") if "twin" in data and data["twin"] is not None else uuid.UUID(data.get("twinId")) if "twinId" in data and data["twinId"] is not None else None,
            edge_device_id=data.get("edgeDeviceId"),
            status=ExecutionStatus(str(data["status"])) if "status" in data else None,
            pipeline_image_id=data.get("pipelineImageId"),
            queued_date=int(data.get("queuedDate")) if "queuedDate" in data and data["queuedDate"] is not None else None,
            started_date=int(data.get("startedDate")) if "startedDate" in data and data["startedDate"] is not None else None,
            waiting_time=int(data.get("waitingTime")) if "waitingTime" in data and data["waitingTime"] is not None else None,
            execution_time=int(data.get("executionTime")) if "executionTime" in data and data["executionTime"] is not None else None,
            trigger_id=uuid.UUID(data["executionTriggerId"]) if "executionTriggerId" in data and data["executionTriggerId"] is not None else None,
            version=data.get("version"),
            properties=properties,
            plots=_plots,
            message=data.get("message"),
            messages=data.get("messages"),
            timestamp=int(data["timestamp"]) if "timestamp" in data and data["timestamp"] is not None else None,
            level=data.get("level")
        )

    def __init__(self,
                 execution_id: uuid.UUID = None,
                 experiment: Optional[Union[str, uuid.UUID, Experiment]] = None,
                 pipeline: Optional[Union[str, uuid.UUID, Pipeline]] = None,
                 twin: Optional[Union[str, uuid.UUID, Twin]] = None,
                 template: Optional[Union[str, uuid.UUID, Template]] = None,
                 registration: Optional[Union[str, uuid.UUID, TwinRegistration]] = None,
                 edge_device_id: str = None,
                 status: ExecutionStatus = None,
                 pipeline_image_id: str = None,
                 timestamp: int = None,
                 started_date: int = None,
                 queued_date: int = None,
                 waiting_time: int = None,
                 execution_time: int = None,
                 trigger_id: uuid.UUID = None,
                 version: str = None,
                 properties: dict = None,
                 plots: list = None,
                 message: str = None,
                 messages: list = None,
                 level: str = None,
                 ):

        # labels
        self._experiment = None
        self._experiment_id = None
        self._experiment_key = None
        self.experiment = experiment

        self._pipeline = None
        self._pipeline_id = None
        self._pipeline_key = None
        self.pipeline = pipeline

        self._twin = None
        self._twin_id = None
        self._twin_key = None
        self.twin = twin

        self._template = None
        self._template_id = None
        self._template_key = None
        self.template = template

        self._registration = None
        self._registration_id = None
        self.registration = registration

        self.edge_device_id = edge_device_id
        self.status = status
        self.level = level

        # values
        if execution_id is None:
            execution_id = uuid.uuid4()
        self.execution_id = execution_id
        self.pipeline_image_id = pipeline_image_id
        self.timestamp = timestamp
        self.started_date = started_date
        self.queued_date = queued_date
        self.waiting_time = waiting_time
        self.execution_time = execution_time
        self.trigger_id = trigger_id
        self.version = version
        if version is None:
            self.version = f'{sys.version_info.major}.{sys.version_info.minor}'

        # non-persistent data
        if properties is None:
            properties = {}
        self.properties = properties
        self.models = []
        if plots is not None:
            self.plots = plots
        else:
            self.plots = []
        self.dataframes = []
        self.message = message
        if messages is not None:
            self.messages = messages
        else:
            self.messages = []

    @property
    def experiment(self):
        """
        get experiment
        :return: object or key or id
        """
        if self._experiment is not None:
            return self._experiment
        if self._experiment_key is not None:
            return self._experiment_key
        return self._experiment_id

    @experiment.setter
    def experiment(self, value: Optional[Union[str, uuid.UUID, Experiment , None]]):
        """
        set experiment - accept key/uuid or a Experiment
        """
        if value is None:
            self._experiment = None
            self._experiment_id = None
            self._experiment_key = None
            return

        if isinstance(value, Experiment):
            self._experiment = value
            self._experiment_key = value.key
            self._experiment_id = value.experiment_id
            return

        if isinstance(value, uuid.UUID):
            self._experiment = None
            self._experiment_id = value
            return

        if isinstance(value, str):
            self._experiment = None
            self._experiment_key = value
            return

        raise TypeError(f"Unsupported pipeline type: {type(value)}")

    def get_experiment_id(self):
        if self._experiment is not None:
            return self._experiment.experiment_id
        else:
            return self._experiment_id

    def get_experiment_key(self):
        if self._experiment is not None:
            return self._experiment.key
        else:
            return self._experiment_key

    @property
    def pipeline(self):
        """
        get pipeline
        :return: object or key or id
        """
        if self._pipeline is not None:
            return self._pipeline
        if self._pipeline_key is not None:
            return self._pipeline_key
        return self._pipeline_id

    @pipeline.setter
    def pipeline(self, value: Optional[Union[str, uuid.UUID, Pipeline , None]]):
        """
        set pipeline - accept key/uuid or a Pipeline
        """
        if value is None:
            self._pipeline = None
            self._pipeline_id = None
            self._pipeline_key = None
            return

        if isinstance(value, Pipeline):
            self._pipeline = value
            self._pipeline_key = value.key
            self._pipeline_id = value.pipeline_id
            return

        if isinstance(value, uuid.UUID):
            self._pipeline = None
            self._pipeline_id = value
            return

        if isinstance(value, str):
            self._pipeline = None
            self._pipeline_key = value
            return

        raise TypeError(f"Unsupported pipeline type: {type(value)}")

    def get_pipeline_id(self):
        if self._pipeline is not None:
            return self._pipeline.pipeline_id
        else:
            return self._pipeline_id

    def get_pipeline_key(self):
        if self._pipeline is not None:
            return self._pipeline.key
        else:
            return self._pipeline_key

    @property
    def twin(self):
        """
        get twin
        :return: object or key or id
        """
        if self._twin is not None:
            return self._twin
        if self._twin_key is not None:
            return self._twin_key
        return self._twin_id

    @twin.setter
    def twin(self, value: Optional[Union[str, uuid.UUID, Twin , None]]):
        """
        set twin - accept key/uuid or a Twin
        """
        if value is None:
            self._twin = None
            self._twin_id = None
            self._twin_key = None
            return

        if isinstance(value, Twin):
            self._twin = value
            self._twin_key = value.hardware_id
            self._twin_id = value.twin_id
            return

        if isinstance(value, uuid.UUID):
            self._twin = None
            self._twin_id = value
            return

        if isinstance(value, str):
            self._twin = None
            self._twin_key = value
            return

        raise TypeError(f"Unsupported twin type: {type(value)}")

    def get_twin_id(self):
        if self._twin is not None:
            return self._twin.twin_id
        else:
            return self._twin_id

    def get_twin_key(self):
        if self._twin is not None:
            return self._twin.hardware_id
        else:
            return self._twin_key

    @property
    def template(self):
        """
        get template
        :return: object or key or id
        """
        if self._template is not None:
            return self._template
        if self._template_key is not None:
            return self._template_key
        return self._template_id

    @template.setter
    def template(self, value: Optional[Union[str, uuid.UUID, Template , None]]):
        """
        set template - accept key/uuid or a Template
        """
        if value is None:
            self._template = None
            self._template_id = None
            self._template_key = None
            return

        if isinstance(value, Template):
            self._template = value
            self._template_key = value.key
            self._template_id = value.template_id
            return

        if isinstance(value, uuid.UUID):
            self._template = None
            self._template_id = value
            return

        if isinstance(value, str):
            self._template = None
            self._template_key = value
            return

        raise TypeError(f"unsupported template type: {type(value)}")

    def get_template_id(self):
        if self._template is not None:
            return self._template.template_id
        else:
            return self._template_id

    def get_template_key(self):
        if self._template is not None:
            return self._template.key
        else:
            return self._template_key

    @property
    def registration(self):
        """
        get registration
        :return: object or id
        """
        if self._registration is not None:
            return self._registration
        return self._registration_id

    @registration.setter
    def registration(self, value: Optional[Union[str, uuid.UUID, TwinRegistration , None]]):
        """
        set template - accept uuid or a TwinRegistration
        """
        if value is None:
            self._registration = None
            self._registration_id = None
            return

        if isinstance(value, TwinRegistration):
            self._registration = value
            self._registration_id = value.twin_registration_id
            return

        if isinstance(value, uuid.UUID):
            self._registration = None
            self._registration_id = value
            return

        raise TypeError(f"Unsupported registration type: {type(value)}")

    def get_registration_id(self):
        if self._registration is not None:
            return self._registration.twin_registration_id
        else:
            return self._registration

    def get_labels(self) -> dict:
        labels = {
            "experiment": self.get_experiment_key(),
            "pipeline": self.get_pipeline_key(),
            "template": self.get_template_key(),
            "twin": self.get_twin_key(),
            "edgeDeviceId": self.edge_device_id,
            "status": self.status.value if hasattr(self.status, 'value') else self.status,
            "level": self.level
        }
        return {k: v for k, v in labels.items() if v is not None}

    def get_values(self) -> dict:
        data = {
            "id": str(self.execution_id) if self.execution_id else None,
            "pipelineImageId": self.pipeline_image_id,
            "timestamp": self.timestamp,
            "startedDate": self.started_date,
            "queuedDate": self.queued_date,
            "waitingTime": self.waiting_time,
            "executionTime": self.execution_time,
            "executionTriggerId": str(self.trigger_id) if self.trigger_id else None,
            "version": self.version,
            "message": self.message,
            "messages": self.messages or None
        }
        if self.plots is not None and len(self.plots) > 0:
            data["plots"] = []
            for plot in self.plots:
                data["plots"].append(plot.to_json(target="logs"))
        return {k: v for k, v in data.items() if v is not None}

    def to_dict(self) -> dict:
        """
        convert the log to a dictionary
        - don't forget to drop properties if needed
        :return: dict of the execution log
        """
        non_values = {
            "properties": self.properties or {}
        }
        return self.get_labels() | self.get_values() | non_values